# !/usr/bin/env python
#  -*- coding: UTF-8 -*-

__version__ = "0.7"  # LATEST? =>  https://pypi.org/project/dimcli/
__copyright__ = "CopyRight (C) 2018-2020 by Michele Pasin"
__license__ = "MIT"
__author__ = "Michele Pasin"
__author_email__ = "michele dot pasin at gmail dot com"

USAGE = "dimcli [options]"
VERSION = "v" + __version__
AGENT = "%s/%s" % (__name__, __version__)



#!/usr/bin/python
# -*- coding: utf-8 -*-


"""
simple client from new oauth API

PRIVATE FILE
"""

import click 
import requests



@click.command()
@click.argument('query', nargs=-1)
def main(query=1):
    """
    Simple test script for new API based on Readcube

    From command line:

    export endpoint="https://sandbox-api.dimensions.ai/"
    export DSL_TOKEN="$(curl -X GET "https://$endpoint/token?api_key=8cGaPHPQfxjmMR5NcnWCwzwjHIleGrTz")"
    curl -X POST https://$endpoint/api/query -H "authorization: Bearer $DSL_TOKEN" -d 'search publications return publications'

    """
    if not query: 
        q = "search publications return publications"
    else:
        q = " ".join([x for x in query])
    

    if True: #analytics 
        api_key = "8cGaPHPQfxjmMR5NcnWCwzwjHIleGrTz"
        endpoint = "sandbox-api.dimensions.ai"
    else: #CRIS
        api_key = "uTQEFiYYeBzLSBGZr39zGBhaxuAiLkHp"
        endpoint = "sandbox-cris-api.dimensions.ai"

    # get token from key    

    url_token = f"https://{endpoint}/token?api_key={api_key}"
    response = requests.get(url_token)
    response.raise_for_status()
    token = response.text.strip("\n") # NOTE token has newline char at the end
    # print("Token:",  token)
    
    # query using token
    
    headers = {'Authorization': "Bearer " + token}
    url_query = f"https://{endpoint}/api/query"
    response = requests.post(url_query, data=q, headers=headers)
    
    try:
        res_json = response.json()
        print(res_json)
    except:
        print('Unexpected error. JSON could not be parsed.')
        print("Raw Response:",  response)



if __name__ == '__main__':
    main()





"""
NOTE this module is deprecated and will be removed in future versions.

Use instead dimcli.utils 
"""


from . import *
from .core.utils import chunks_of, normalize_key, dimensions_url, google_url, save2File, exists_key_in_dicts_list, dsl_escape



def dslquery(query_string):
    """shortcut for running a query - meant to be used only within interactive computing environments
    """
    from .core.auth import is_logged_in
    if is_logged_in():
        dsl = Dsl()
        res = dsl.query(query_string, verbose=True)
        return res


def dslquery_json(query_string):
    """shortcut for backward compatibility 
    Same as above but returns raw JSON instead of Api.DslDataset object

    Pattern: `from dimcli.shortcuts import dslquery_json as dslquery`
    """
    from .core.auth import is_logged_in
    if is_logged_in():
        dsl = Dsl()
        return dsl.query(query_string).json


def dslqueryall(query_string):
    """shortcut for running a loop query - meant to be used only within interactive computing environments
    NOTE: this requires the file-based credentials file set up.
    """
    from .core.auth import is_logged_in
    if is_logged_in():
        dsl = Dsl()
        return dsl.query_iterative(query_string)




def extract_concepts(text, with_scores=True, as_df=True):
    """wrapper for the `extract_concepts` function
    https://docs.dimensions.ai/dsl/functions.html#function-extract-concepts
    """
    from .core.auth import is_logged_in
    if is_logged_in():
        dsl = Dsl()
        _score = 'true' if with_scores else 'false'
        if as_df:
            return dsl.query(f"""extract_concepts("{text}", return_scores={_score})""").as_dataframe()
        else:
            return dsl.query(f"""extract_concepts("{text}", return_scores={_score})""")




def extract_grants(grant_number, fundref="", funder_name=""):
    """wrapper for the `extract_grants` function
    https://docs.dimensions.ai/dsl/functions.html#function-extract-grants
    NOTE either fundref or funder_name needs to be provided
    """
    from .core.auth import is_logged_in
    if is_logged_in():
        dsl = Dsl()
        if fundref:
            return dsl.query(f"""extract_grants(grant_number="{grant_number}", fundref="{fundref}")""")
        else:
            return dsl.query(f"""extract_grants(grant_number="{grant_number}", funder_name="{funder_name}")""")



def extract_classification(title, abstract, system="", verbose=True):
    """wrapper for the `classify` function
    https://docs.dimensions.ai/dsl/functions.html#function-classify

    `system` must be an acronym from the supported classification systems:

    * Fields of Research (FOR)
    * Research, Condition, and Disease Categorization (RCDC)
    * Health Research Classification System Health Categories (HRCS_HC)
    * Health Research Classification System Research Activity Classifications (HRCS_RAC)
    * Health Research Areas (HRA)
    * Broad Research Areas (BRA)
    * ICRP Common Scientific Outline (ICRP_CSO)
    * ICRP Cancer Types (ICRP_CT)
    * Units of Assessment (UOA)
    * Sustainable Development Goals (SDG)

    """
    from .core.auth import is_logged_in
    classifications = ["FOR", "RCDC", "HRCS_HC", "HRCS_RAC", "HRA", "BRA", "ICRP_CSO", "ICRP_CT", "UOA", "SDG"]
    if is_logged_in():
        dsl = Dsl()
        if system:
            return dsl.query(f"""classify(title="{dsl_escape(title)}", 
                                        abstract="{dsl_escape(abstract)}", 
                                        system="{system}")""")
        else:
            if verbose: print(f"""No system provided, using all known systems ({len(classifications)} queries). Warning: This may lead to 'too many API queries' errors.""")
            d = {}
            for classifier in classifications:
                new = dsl.query(f"""classify(title="{dsl_escape(title)}", 
                                        abstract="{dsl_escape(abstract)}", 
                                        system="{classifier}")""").json
                d.update(new)
            return d


#!/usr/bin/env python
"""
Autocompletion example.

Press [Tab] to complete the current word.
- The first Tab press fills in the common part of all completions
    and shows all the completions. (In the menu)
- Any following tab press cycles through all the possible completions.
"""
from __future__ import unicode_literals

from prompt_toolkit.completion import Completion, Completer
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.shortcuts import CompleteStyle
from prompt_toolkit import PromptSession
from prompt_toolkit.styles import Style

from .autocompletion import *
from .history import *
from .key_bindings import *
from .lexer import *



# from prompt_toolkit import prompt   #using session instead


import click
import json
import sys
import os
import time
import requests

from ..core.api import Dsl
from ..core.auth import USER_HISTORY_FILE, USER_EXPORTS_DIR, do_global_login, get_connection
from ..core.dsl_grammar import *
from ..core.utils import *





HELP_MESSAGE =  """COMMANDS LIST
====================
All special commands start with '.'
----
>>> help: show this help message
----
>>> <tab>:  autocomplete. 
----
>>> .docs: print out documentation for DSL data objects.  
>>> .export_as_json: save results from last query as JSON file.  
>>> .export_as_csv: save results from last query as CSV file.  
>>> .export_as_html: save results from last query as HTML file. 
>>> .export_as_bar_chart: save results from last query as Plotly bar chart. 
>>> .export_as_jupyter: save results from last query as Jupyter notebook. 
>>> .show [optional: N]: print N results from last query, trying to build URLs for objects. Default N=10.
>>> .json_compact: print results of last query as single-line JSON. 
>>> .json_full: print results of last query as formatted JSON.
>>> .url: resolve a Dimensions ID into a public URL.
----
>>> <Ctrl-o>: search docs online. 
>>> <Ctrl-c>: abort query.
>>> <Ctrl-d>: exit console.
----
>>> quit: exit console
====================
*ABOUT AUTOCOMPLETE*
Including a space between query elements (= keywords and operators) leads to better autocomplete results."""

WELCOME_MESSAGE = "Welcome! Type help for more info."
# WELCOME_MESSAGE = "Welcome! Type help for more info. Ready to query endpoint: %s"



#
#
# DIMENSIONS QUERY AND DATA HANDLING
#
#


class DslResultsBuffer(object):
    current_json = ""
    current_query = ""

    def save(self, json_data, query):
        self.current_json = json_data
        self.current_query = query
        self.is_recording = False

    def retrieve(self):
        return (self.current_json, self.current_query)



class CommandsManager(object):

    def __init__(self, dslclient, databuffer):
        self.dsl = dslclient
        self.bf = databuffer

    def handle(self, text):
        "process text and delegate"
        if text.replace("\n", "").strip().startswith(".show") or text.replace("\n", "").strip().startswith(".json"):
            self.show(text.replace("\n", "").strip())

        elif text.replace("\n", "").strip().startswith(".export"):
            self.export(text.replace("\n", "").strip())

        elif text.replace("\n", "").strip().startswith(".docs"):
            self.docs_full(text.replace("\n", "").strip())

        elif text.replace("\n", "").strip().startswith(".url"):
            self.url_resolver(text.replace("\n", "").strip())

        elif text.replace("\n", "").strip().startswith(".record"):
            self.record_notebook(text.replace("\n", "").strip())

        else:
            return self.query(text)


    def url_resolver(self, text):
        """
        turn an ID into a Dimensions URL - print out results
        """
        text = text.replace(".url", "").strip()
        if len(text) > 0:
            print_dimensions_url(text)

    def query(self, text):
        """main procedure after user query dsl"""
        # lazy complete
        text = line_add_lazy_return(text)
        text = line_add_lazy_describe(text)
        click.secho("You said: %s" % text, fg="black", dim=True)
        # RUN QUERY
        res = self.dsl.query(text)
        # errors info gets printed out by the API by default
        if not "errors" in res.json.keys():
            if "_warnings" in res.json.keys():
                click.secho("WARNINGS [{}]".format(len(res.json["_warnings"])), fg="red")
                # print("WARNINGS [{}]".format(len(res.json["_warnings"])))
                print("\n".join([s for s in res.json["_warnings"]]))
                click.secho("---", dim=True)
            # search query
            if text.strip().startswith("search"):
                print_json_stats(res, text)
                if self.bf: self.bf.save(res.json, text)
                if True:
                    click.secho("---", dim=True)
                    preview_results(res.json, maxitems=5)
                return res  # 2019-03-31
            # describe queries and other functions: just show the data
            else:
                if self.bf: self.bf.save(res.json, text)
                click.secho("---", dim=True)
                print_json_full(res.json)


    def docs_full(self, text):
        """
        print out docs infos from 'describe' API
        """
        text = text.replace(".docs", "").split()
        if len(text) > 0:
            if text[0] in G.entities():
                res = self.dsl.query(f"describe entity {text[0]}")
            else:
                res = self.dsl.query(f"describe source {text[0]}")
            if "errors" in res.json.keys():
                print(res.json["errors"])
                return 
            # show all fields 
            click.secho("=====\nFIELDS")
            for x in sorted(res.json['fields']):
                d = res.json['fields'][x].get('description') or  ""
                typ = res.json['fields'][x].get('type', "")
                ise = res.json['fields'][x].get('is_entity', "")
                isfa = res.json['fields'][x].get('is_facet', "")
                isfi = res.json['fields'][x].get('is_filter', "")
                if isfa:
                    infos = f"[type={typ}, filter={isfi}, facet={isfa}]"
                else:
                    infos = f"[type={typ}, filter={isfi}]"
                click.echo(click.style(x, bold=True) + " " + click.style(infos, dim=True) + \
                    " " + click.style(d))
            if 'metrics' in res.json:
                click.secho("=====\nMETRICS")
                for x in res.json['metrics']:
                    d = res.json['metrics'][x]['description'] or "no description"
                    click.echo(click.style(x, bold=True) + " " + click.style(d, dim=True))
            if 'fieldsets' in res.json:
                click.secho("=====\nFIELDSETS")
                f = res.json['fieldsets']
                if f:
                    click.secho(", ".join([x for x in f]), bold=True)
            if 'search_fields' in res.json:
                click.secho("=====\nSEARCH FIELDS")
                f = res.json['search_fields']
                if f:
                    click.secho(", ".join([x for x in f]), bold=True)

        else:
            print("Please specify a source or entity.")

    def record_notebook(self, text, rows=10):
        """
        2020-06-01: DEPRECATED TODO remove 
        Sustituted with .export function
        
        Take the last 10 (default) rows from the history, and create a new python notebook with them. 
        Saves in usual location.

        Based on https://gist.github.com/fperez/9716279

        @TODO handle is history rows are not enough! 
        @ TODO Move imports
        """
        from itertools import islice
        import nbformat as nbf
        init_exports_folder(USER_EXPORTS_DIR)
        history=SelectiveFileHistory(USER_HISTORY_FILE)

        rows_data = []
        for item in islice(history.load_history_strings(), rows):
            rows_data += [item]
        rows_data.reverse()

        nb = nbf.v4.new_notebook()
        this_time = time.strftime("%Y.%m.%d h%H:%M:%S")

        text = f"""# Dimensions API Queries Export\n### {this_time} \nThis notebook was generated using [Dimcli](https://github.com/digital-science/dimcli/) - the Dimensions API CLI."""
    
        nb['cells'] = [nbf.v4.new_markdown_cell(text)]

        setup = """import dimcli\ndimcli.login()"""
        nb['cells'] += [nbf.v4.new_code_cell(setup)]

        for code in rows_data:
            nb['cells'] += [nbf.v4.new_code_cell("%dsldf " + code)]

        filename = time.strftime(f"dsl_export_%Y-%m-%d_%H-%M-%S.ipynb")
        nbf.write(nb, USER_EXPORTS_DIR + filename)
        import subprocess
        subprocess.run(['open', USER_EXPORTS_DIR + filename], check=True)
        print("Exported: ", "%s%s" % (USER_EXPORTS_DIR, filename))


    def export(self, text):
        """
        save results of a query to a file
        """
        init_exports_folder(USER_EXPORTS_DIR)
        if self.bf: 
            jsondata, query = self.bf.retrieve()
        else:
            jsondata, query = None, None
        if not jsondata:
            print("Nothing to export - please run a search first.")
            return
        CONNECTION = get_connection()
        api_endpoint = CONNECTION['url']
        # cases
        if text == ".export_as_html":
            export_json_html(jsondata, query, api_endpoint, USER_EXPORTS_DIR)

        elif text == ".export_as_csv":
            export_json_csv(jsondata, query, USER_EXPORTS_DIR)

        elif text == ".export_as_json":
            export_json_json(jsondata, query, USER_EXPORTS_DIR)

        elif text == ".export_as_bar_chart":
            export_as_bar_chart(jsondata, query, USER_EXPORTS_DIR)

        elif text == ".export_as_jupyter":
            export_as_jupyter(jsondata, query, USER_EXPORTS_DIR)


    def show(self, text):
        """
        show results of a query
        """
        DEFAULT_NO_RECORDS = 10
        # text = text.replace(".show", "").strip()
        text = text.strip()

        if self.bf: 
            jsondata, query = self.bf.retrieve()
        else:
            jsondata, query = None, None
        if not jsondata:
            print("Nothing to show - please run a search first.")
            return
        # cases
        if text == ".json_compact":
            print_json_compact(jsondata)
        elif text == ".json_full":
            print_json_full(jsondata)
        else:
            # must be a simple ".show" + X command
            try:
                no = text.replace(".show", "").strip()
                slice_no = int(no)
            except ValueError:
                slice_no = DEFAULT_NO_RECORDS
            preview_results(jsondata, maxitems=slice_no)




#
#
# MAIN CLI
#
#


def run(instance="live"):
    """
    run the repl
    """

    try:
        do_global_login(instance=instance)
        CLIENT = Dsl(verbose=False)
    except requests.exceptions.HTTPError as err:
        print(err)
        sys.exit(1)

    # dynamically retrieve dsl version 
    click.secho(WELCOME_MESSAGE)
    try:
        _info = CLIENT.query("describe version")['release']
    except:
        _info = "not available"
    click.secho(f"Using endpoint: {CLIENT._url} - DSL version: {_info}", dim=True)

    # history
    session = PromptSession(history=SelectiveFileHistory(USER_HISTORY_FILE))

    databuffer = DslResultsBuffer()

    # REPL loop
    while True:
        try:
            text = session.prompt(
                "\n> ",
                default="",  # you can pass a default text to begin with
                completer=CleverCompleter(),
                complete_style=CompleteStyle.MULTI_COLUMN,
                # validator=BasicValidator(),
                # validate_while_typing=False,
                multiline=False,
                complete_while_typing=True,
                lexer=BasicLexer(),
                key_bindings=bindings,
            )
        except KeyboardInterrupt:
            continue  # Control-C pressed. Try again.
        except EOFError:
            break  # Control-D pressed.
        except Exception as e:
            print(e)
            sys.exit(0)  
        else:
            if text.strip() == "":
                continue
            elif text == "quit":
                break
            elif text == "help":
                click.secho(HELP_MESSAGE, dim=True)
                continue
            try:
                cm = CommandsManager(CLIENT,databuffer)
                cm.handle(text)
            except Exception as e:
                print(e)
                continue
    print("GoodBye!")


if __name__ == "__main__":
    run()



# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Unit tests for Dimcli - Dataframes
"""

from __future__ import print_function

import unittest, os, sys, click
import configparser

from .. import *
from ..core.auth import USER_CONFIG_FILE_PATH
from ..core.utils import dimensions_url
from ..shortcuts import *


class TestDataframes(unittest.TestCase):

    """
    Tests  
    """

    click.secho("**TESTS**", fg="red")
    login(instance="live")
    d = Dsl()

    def test_001(self):
        click.secho("\nTEST 001: Generic Dataframes.", fg="green")
        # ----
        click.secho("search publications for \"mercedes\" return year", fg="green")
        res = dslquery("""search publications for \"mercedes\" return year""")
        click.secho("Query results: ", fg="magenta")
        print(" ==> res.json.keys(): ", res.json.keys())
        print(" ==> len(res): ", len(res))
        print(" ==> res.stats: ", res.stats)
        print(" ==> res.as_dataframe(): ", res.as_dataframe())
        print(" ==> res.as_dataframe('year'): ", res.as_dataframe('year'))
        print(" ==> res.as_dataframe('XXX'): [FAILS] ", res.as_dataframe('XXX'))


    def test_002(self):
        click.secho("\nTEST 002: Authors and Affiliations Dataframes.", fg="green")
        # ----
        click.secho("Testing as_dataframe_authors with legacy `author_affiliations` field: ", fg="magenta")
        res = dslquery("""search publications for \"CRISPR\" return publications limit 5""")
        print("Testing as_dataframe_authors on PUBLICATIONS data: ")
        print(" ==> res.as_dataframe_authors(): ", res.as_dataframe_authors())
        print(" ==> res.as_dataframe_authors_affiliations(): ", res.as_dataframe_authors_affiliations())
        # ----
        click.secho("Testing as_dataframe_authors with new `authors` field: ", fg="magenta")
        res = dslquery("""search publications for \"CRISPR\" return publications[doi+id+authors] limit 5""")
        click.secho("Testing as_dataframe_authors on PUBLICATIONS data: ", fg="magenta")
        print(" ==> res.as_dataframe_authors(): ", res.as_dataframe_authors())
        print(" ==> res.as_dataframe_authors_affiliations(): ", res.as_dataframe_authors_affiliations())
        # ----
        click.secho("Testing as_dataframe_authors on results that have both fields: ", fg="magenta")
        res = dslquery("""search publications for \"CRISPR\" return publications[basics+authors] limit 5""")
        click.secho("Testing as_dataframe_authors on PUBLICATIONS data: ", fg="magenta")
        print(" ==> res.as_dataframe_authors(): ", res.as_dataframe_authors())
        print(" ==> res.as_dataframe_authors_affiliations(): ", res.as_dataframe_authors_affiliations())
        # ----
        click.secho("Testing as_dataframe_authors on FOR data (FAIL): ", fg="magenta")
        res = dslquery("""search publications for \"CRISPR\" return FOR""")
        print(" ==> res.as_dataframe(): ", res.as_dataframe())
        print(" ==> res.as_dataframe_authors(): [=>fails and issue a warning]", res.as_dataframe_authors())
        print(" ==> res.as_dataframe_authors_affiliations(): [=>fails and issue a warning]", res.as_dataframe_authors_affiliations())

    def test_003(self):
        click.secho("\nTEST 003: Grants: funders and investigators Dataframes.", fg="green")
        # ----
        click.secho("Testing as_dataframe_funders on Grants data: ", fg="magenta")
        res= dslquery("""search grants return grants[basics+investigator_details]""")
        print(" ==> res.as_dataframe_funders(): ", res.as_dataframe_funders())
        print(" ==> res.as_dataframe_investigators(): ", res.as_dataframe_investigators())
        click.secho("Completed test succesfully", fg="green")


    def test_004(self):
        click.secho("\nTEST 004: UTILS: normalize_key", fg="green")
        
        # ----
        res = dslquery("""search publications where category_for is empty and journal is empty return publications[doi+category_for+journal] limit 1000""")
        click.secho("Query results for standard query: ", fg="magenta")
        print(" ==> res['stats']: ", res['stats'])
        print(" ==> len(res['publications']): ", len(res['publications']))
        # check predicates
        print(" ==> len([x for x in res.publications if 'category_for' in x]): ", len([x for x in res.publications if 'category_for' in x ]))
        print(" ==> len([x for x in res.publications if 'journal' in x]): ", len([x for x in res.publications if 'journal' in x]))
        click.secho("Now Normalizing the category_for key...", fg="magenta")
        normalize_key("category_for", res.publications, {})
        click.secho("Now Normalizing the JOURNAL key...", fg="magenta")
        normalize_key("journal", res.publications, [])
        print(" ==> len([x for x in res.publications if 'category_for' in x]): ", len([x for x in res.publications if 'category_for' in x]))
        print(" ==> len([x for x in res.publications if 'journal' in x]): ", len([x for x in res.publications if 'journal' in x]))
        # ----
        click.secho("Completed test succesfully", fg="green")

    def test_005(self):
        # ----
        click.secho("\nTEST 005: Concepts extraction.", fg="green")
        # ----
        click.secho("Testing as_dataframe_concepts on Publications data: ", fg="magenta")
        res= dslquery("""search publications for "graphene" where year=2019 return publications[id+concepts+year+title+doi+journal] limit 1000""")
        concepts = res.as_dataframe_concepts()
        print(" ==> res.as_dataframe_concepts(): ", concepts)
        concepts.info()
        # ----
        click.secho("Testing as_dataframe_concepts on Publications data - using SCORES: ", fg="magenta")
        res= dslquery("""search publications for "graphene" where year=2019 return publications[id+concepts_scores+year+title+doi+journal] limit 1000""")
        concepts = res.as_dataframe_concepts()
        print(" ==> res.as_dataframe_concepts(): ", concepts)
        concepts.info()
        # ----
        click.secho("Testing as_dataframe_concepts on Grants data: ", fg="magenta")
        res= dslquery("""search grants for "graphene" where active_year = 2019 return grants[basics+concepts] limit 500""")
        concepts = res.as_dataframe_concepts()
        print(" ==> res.as_dataframe_concepts(): ", concepts)
        concepts.info()
        click.secho("Completed test succesfully", fg="green")

if __name__ == "__main__":
    unittest.main()



# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Unit tests for Dimcli  - logins
"""

from __future__ import print_function

import unittest, os, sys, click
import configparser

from .. import *
from ..core.auth import USER_CONFIG_FILE_PATH
from ..core.utils import dimensions_url
from ..shortcuts import *


class TestOne(unittest.TestCase):

    """
    Tests  
    """

    click.secho("**TESTS**", fg="red")

    def test_001(self):
        click.secho("\nTEST 001: load Dimcli using file-based credentials and verbose mode", fg="green")
        # ----
        click.secho("Login... verbose=True", fg="magenta")
        login()
        d = Dsl()
        res = d.query("search publications where year=2018 return publications")
        print("Query BATCH results: ", res.count_batch)
        click.secho("Logout... verbose=True", fg="magenta")
        logout()
        # ----
        click.secho("Login... verbose=False", fg="magenta")
        login(verbose=False)
        d = Dsl()
        res = d.query("search publications where year=2018 return publications")
        print("Query BATCH results: ", res.count_batch)
        click.secho("Logout... verbose=False", fg="magenta")
        logout()
        # ----
        click.secho("Completed test succesfully", fg="green")

    def test_002(self):
        click.secho("\nTEST 002: load Dimcli by passing credentials explicitly.", fg="green")
        # ----
        # get credentials from file as strings
        config = configparser.ConfigParser()
        config.read(os.path.expanduser(USER_CONFIG_FILE_PATH))
        section = config['instance.live' ]
        USER = section['login']
        PSW = section['password']
        login(USER, PSW)
        d = Dsl()
        res = d.query("search publications where year=2018 return publications")
        # print("Query results: ", res.keys_and_count())
        # ----
        logout()
        click.secho("\n--------\nCompleted test succesfully", fg="green")

    def test_002_1(self):
        click.secho("\nTEST 002-1: Retain login info and force new login.", fg="green")
        # ----
        login(instance="live")
        d = Dsl()
        click.secho(""" Dsl(instance="live"): ==> url=""" + d._url, fg="magenta")
        res = d.query("""search publications where authors="Pasin" return publications""")
        print(" ==> res.json.keys(): ", res.json.keys())
        logout()
        login(instance="test")
        d = Dsl()
        click.secho(""" d.login(instance="test"): ==> url=""" + d._url, fg="magenta")
        res = d.query("""search publications where authors="Pasin" return publications""")
        print(" ==> res.json.keys(): ", res.json.keys())
        logout()
        # ----
        click.secho("\n--------\nCompleted test succesfully", fg="green")


    def test_003(self):
        click.secho("\nTEST 003: Login using key-based authentication.", fg="green")
        # ----
        logout()
        login(instance="key-test")
        d = Dsl()
        click.secho(""" Dsl(instance="key-test"): ==> url="""+ d._url, fg="magenta")
        res = d.query("""search publications where authors="Pasin" return publications""")
        print(" ==> res.json.keys(): ", res.json.keys())
        logout()
        # ----
        click.secho("\n--------\nCompleted test succesfully", fg="green")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from prompt_toolkit.validation import Validator, ValidationError

#
# VALIDATOR
#
#
# UNUSED FOR NOW


class BasicValidator(Validator):
    def validate(self, document):
        text = document.text

        if text and "return" not in text:

            raise ValidationError(
                message="A query must include a return statement",
                # cursor_position=i
            )



# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Unit tests for Dimcli 
"""

from __future__ import print_function

import unittest, os, sys, click
import configparser

from .. import *
from ..core.auth import USER_CONFIG_FILE_PATH
from ..core.utils import dimensions_url
from ..core.api import DslDataset
from ..shortcuts import *


class TestOne(unittest.TestCase):

    """
    Tests - DSL function wrappers 
    """

    click.secho("**TESTS**", fg="red")
    login(instance="live")
    d = Dsl()

    def test_001(self):
        click.secho("\nTEST 001: Extract concepts.", fg="green")
        # ----
        a = """
BACKGROUND: In order to make further gains in preventing newborn deaths, effective interventions are needed. Ultrasounds and newborn anthropometry are proven interventions to identify preterm birth complications, the leading cause of newborn deaths. The INTERGROWTH-21st global gestational dating and fetal and newborn growth standards prescribe optimal growth in any population. Jacaranda Health in Kenya was the first low-resource health facility to implement the standards and evaluate their feasibility and acceptability. OBJECTIVE: To capture patients' perceptions of ultrasound and newborn care before and during implementation of the INTERGROWTH-21st standards. METHODS: The study was conducted over two years before and during the introduction of the INTERGROWTH-21st standards. Fifty pregnant and/or newly delivered women were selected for in-depth interviews and focus group discussions using convenience and purposive sampling. Interviews were conducted by research assistants using semi-structured guides once in the pre-implementation phase and twice in the implementation phase. Interviews were transcribed, double-coded by two independent researchers and thematically analyzed together. Demographic information was obtained from hospital records. RESULTS: Patients reported being generally satisfied with ultrasound care when providers communicated effectively. Women reported a priority for ultrasound was that it allowed them to feel reassured. However, a clear need for better pre-screening information emerged consistently from patients. Women noted that factors facilitating their choosing to have an ultrasound included ensuring the well-being of the fetus and learning the sex. Barriers included wait times and financial constraints. Patients were generally satisfied with care using the newborn standards. CONCLUSIONS: As the INTERGROWTH-21st standards are implemented worldwide, understanding ways to facilitate implementation is critical. Increased and standardized communication about ultrasound should be provided before the procedure to increase satisfaction and uptake. Considering patient perspectives when integrating new standards or guidelines into routine clinical care will inform effective strategies in care provision, thus improving maternal and newborn health and survival."""
        click.secho("With scores", fg="magenta")
        print(extract_concepts(a, with_scores=True))
        click.secho("Without scores", fg="magenta")
        print(extract_concepts(a, with_scores=False))
        click.secho("Without scores, as dimcli.DslDataset", fg="magenta")
        print(extract_concepts(a, with_scores=False, as_df=False))
        # ----
        click.secho("Completed test succesfully", fg="green")



    def test_002(self):
        click.secho("\nTEST 002: Extract grants.", fg="green")
        # ----
        click.secho("With fundref", fg="magenta")
        print(extract_grants("R01HL117329",  fundref="100000050"))
        click.secho("With funder_name", fg="magenta")
        print(extract_grants("HL117648",  funder_name="NIH"))
        click.secho("Build a dataframe", fg="magenta")
        print(extract_grants("HL117648",  funder_name="NIH").as_dataframe())
        # ----
        click.secho("Completed test succesfully", fg="green")



    def test_003(self):
        click.secho("\nTEST 003: Extract classifications.", fg="green")
        # ----
        title="""Burnout and intentions to quit the practice among community pediatricians: 
        associations with specific professional activities"""
        abstract="""BACKGROUND: Burnout is an occupational disease expressed by loss of mental and physical energy due to prolonged and unsuccessful coping with stressors at work. A prior survey among Israeli pediatricians published in 2006 found a correlation between burnout and job structure match, defined as the match between engagement with, and satisfaction from, specific professional activities. The aims of the present study were to characterize the current levels of burnout and its correlates among community pediatricians, to identify changes over time since the prior survey, and to identify professional activities that may reduce burnout. METHODS: A questionnaire was distributed among pediatricians both at a medical conference and by a web-based survey. RESULTS: Of the 518 pediatricians approached, 238 (46%) responded to the questionnaire. High burnout levels were identified in 33% (95% CI:27-39%) of the respondents. Higher burnout prevalence was found among pediatricians who were not board-certified, salaried, younger, and working long hours. The greater the discrepancy between the engagement of the pediatrician and the satisfaction felt in the measured professional activities, the greater was the burnout level (p < 0.01). The following activities were especially associated with burnout: administrative work (frequent engagement, disliked duty) and research and teaching (infrequent engagement, satisfying activities). A comparison of the engagement-satisfaction match between 2006 and 2017 showed that the discrepancy had increased significantly in research (p < 0.001), student tutoring (P < 0.001), continuing medical education and participation in professional conferences (P = 0.0074), management (p = 0.043) and community health promotion (P = 0.006). A significant correlation was found between burnout and thoughts of quitting pediatrics or medicine (p < 0.001). CONCLUSIONS: Healthcare managers should encourage diversification of the pediatrician's job by enabling greater engagement in the identified anti-burnout professional activities, such as: participation in professional consultations, management, tutoring students and conducting research."""
        click.secho("With FORs", fg="magenta")
        print(extract_classification(title, abstract, "FOR").json)
        click.secho("With UOA", fg="magenta")
        print(extract_classification(title, abstract, "UOA").json)
        click.secho("Build a dataframe", fg="magenta")
        print(extract_classification(title, abstract, "FOR").as_dataframe())
        click.secho("Without specifying a system - use all", fg="magenta")
        print(extract_classification(title, abstract))
        # ----
        click.secho("Completed test succesfully", fg="green")



if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""
Autocompletion.

Press [Tab] to complete the current word.
- The first Tab press fills in the common part of all completions
    and shows all the completions. (In the menu)
- Any following tab press cycles through all the possible completions.
"""

from __future__ import unicode_literals

# NOTE this is compatible also with older versions of prompt-toolkit (eg < 2) so will run in Google Colab or Conda 
from prompt_toolkit.completion import Completion, Completer

from ..core.dsl_grammar import *
from ..core.utils import *

#
# AUTO COMPLETION
#


class CleverCompleter(Completer):
    """
    Goal: complete that helps with Dimensions DSL grammar and predicates 

    info: https://python-prompt-toolkit.readthedocs.io/en/master/pages/reference.html#prompt_toolkit.document.Document
    
    """

    def get_completions(self, document, complete_event):
        #
        word = document.get_word_before_cursor(WORD=True)
        line = document.current_line_before_cursor
        line_minus_current = line.replace(word, "").strip()
        source = line_search_subject(line)
        # DEBUG
        # click.secho("\nAutocomplete running..", dim=True)
        # click.secho("WORD=" + word, dim=True)
        # click.secho("LINE=" + line, dim=True)
        # click.secho("LINE_MINUS_CURRENT=" + line_minus_current, dim=True)

        # line_minus_current = line
        candidates = []

        if not line:  # no letter, all empty
            candidates = G.allowed_starts_dsl_query()

        elif len(line_minus_current) == 0:  # single word, beginning of line
            candidates = G.allowed_starts()

        elif line_minus_current and word.endswith("."):
            entity_facet = line_last_word(line).replace(".", "")
            entity = G.entity_type_for_source_facet(source, entity_facet)
            candidates = G.fields_for_entity_from_source_facet(source, entity_facet)

        elif in_categories_search(line):
            this_category = in_categories_search(line)
            # print(this_category)
            candidates = G.categories(this_category)
 
        elif in_square_brackets(line):
            # https://docs.dimensions.ai/dsl/language.html#return-specific-fields
            # search publications for "bmw" return journal[id + title]"
            test_return_obj = line_last_return_subject(line)
            if test_return_obj == source:
                # print("*" + test_return_obj + "*")
                candidates = G.fields_for_source(test_return_obj)
            elif test_return_obj in G.facets_for_source(source):
                entity = G.entity_type_for_source_facet(source, test_return_obj)
                candidates = G.fields_for_entity_from_source_facet(source, test_return_obj)

        elif line_last_word(line_minus_current) in [".docs"]:
            candidates = G.sources() + G.entities()

        elif line_last_word(line_minus_current) in ["describe"]:
            candidates = G.allowed_starts("describe")

        elif line_last_word(line_minus_current) in ["search"]:
            candidates = G.sources()

        elif line_last_word(line_minus_current) in ["return"]:
            if source in G.sources():
                candidates = G.facets_for_source(source) + [source]

        elif line_search_subject_is_valid(line_minus_current):
            candidates = G.lang_after_search()

        elif line_return_subject_is_valid(line_minus_current):
            test_return_obj = line_last_return_subject(line)
            if test_return_obj == source:
                candidates = [x for x in G.lang_after_return() if x != 'aggregate']
            else:
                candidates = G.lang_after_return()

        elif line_last_word(line_minus_current) == "in":
            candidates = G.search_fields_for_source(source)

        elif line_last_word(line_minus_current) in ["where", "and", "or", "not"]:
            candidates = G.filters_for_source(source)

        elif line_filter_is_partial(line_minus_current):
            candidates = G.lang_filter_operators()

        elif line_for_text_search_inner(line_minus_current):
            candidates = G.lang_text_operators()

        elif line_last_two_words(line_minus_current).startswith("limit"):
            if line_last_return_subject(line) == source:
                candidates = G.lang_after_limit()

        elif line_last_word(line_minus_current) == "aggregate":
            # aggr. can be used only when returning facets!
            return_object = line_search_return(line)
            if return_object in G.facets_for_source(source):
                candidates = G.metrics_for_source(source)

        elif line_last_two_words(line_minus_current) == "sort by":  
            return_object = line_search_return(line)  
            if return_object in G.sources(): 
                # if source, can sort by fields FIXME
                candidates = G.fields_for_source(source) + ['relevance']
            elif return_object in G.facets_for_source(source):
                # if facet, can sort by aggregates metrics if available, otherwise count
                aggreg_object = line_search_aggregates(line)
                if aggreg_object:
                    candidates = [aggreg_object]   
                else:
                    candidates = ['count']          

        elif line_last_three_words(line_minus_current).startswith("sort by"): 
            candidates = G.lang_after_sort_by()

        # IMP following two must go last
        elif line_filter_is_complete(line_minus_current):  
            candidates = G.lang_after_filter()
        elif line_for_text_is_complete(line_minus_current): 
            candidates = G.lang_after_for_text()
        # finally
        else:
            candidates = [] 

        #
        # now build the candidates list
        #

        if line_minus_current and word.endswith("."):
            # print("***" + str(candidates) + "***")
            candidates = sorted([word + x for x in candidates])
            for keyword in candidates:
                yield Completion(
                    keyword, 
                    start_position=-len(word),
                    display=keyword.replace(word, ""),
                    display_meta=build_help_string(keyword.replace(word, ""), entity=entity),
                    )

        elif in_square_brackets(line):
            # print("***" + str(word) + "***")
            candidates = sorted(candidates)
            for keyword in candidates:
                if word.rfind("+") > 0:
                    word = word[word.find("+")+1:]
                else:
                    word = word[word.find("[")+1:]
                if keyword.startswith(word):
                    yield Completion(
                        keyword, 
                        start_position=-len(word),
                        display=keyword,
                        display_meta=build_help_string(keyword, source=source),
                        )      

        elif in_categories_search(line):
            candidates = sorted([word + x for x in candidates])
            for keyword in candidates:
                yield Completion(
                    keyword, 
                    start_position=-len(word),
                    display=keyword.replace(word, ""),
                    display_meta=build_help_string(keyword),
                    )  
     
        else:
            candidates = sorted(candidates)
            for keyword in candidates:
                if keyword.startswith(word):
                    yield Completion(
                        keyword, 
                        start_position=-len(word),
                        display=keyword,
                        display_meta=build_help_string(keyword, source=source),
                        )


def build_help_string(field, source="", entity=""):
    if source:
        return G.desc_for_source_field_enriched(source, field)
    elif entity:
        return G.desc_for_entity_field_enriched(entity, field)
    else:
        return ""




#!/usr/bin/python
# -*- coding: utf-8 -*-


#
#
# SYNTAX_DICT is a dictionary representation of operators and other constants of the DSL language 
#
#

SYNTAX_DICT = {
    'allowed_starts_special_commands': {
        'help' : [],
        '.docs' : [],
        'quit' : [],
        '.show' : [],
        '.json_compact' : [],
        '.json_full' : [],
        '.export_as_html' : [],
        '.export_as_csv' : [],
        '.export_as_json' : [],
        '.export_as_bar_chart' : [],
        '.export_as_jupyter' : [],
        '.record_notebook' : [],
        '.url' : [],
    },
    'allowed_starts_dsl_query': {
        'search': [],
        'describe': [ 'version', 'source', 'entity', 'schema'],
        'check_researcher_ids': [],
        'classify': [],
        'extract_grants': [],
        'extract_concepts': [],
        'extract_affiliations': [],
    },
    'dimensions_urls' : {
        'publications' : 'https://app.dimensions.ai/details/publication/',
        'grants' : 'https://app.dimensions.ai/details/grant/',
        'patents' : 'https://app.dimensions.ai/details/patent/',
        'policy_documents' : 'https://app.dimensions.ai/details/policy_documents/',
        'clinical_trials' : 'https://app.dimensions.ai/details/clinical_trial/',
        'datasets' : 'https://app.dimensions.ai/details/data_set/',
        'researchers' : 'https://app.dimensions.ai/discover/publication?and_facet_researcher=',
        'organizations' : 'https://app.dimensions.ai/discover/publication?and_facet_research_org=',
    },
    'dimensions_object_id_patterns' : {
        'publications' : 'pub.',
        'grants' : 'grant.',
        # 'patents' : 'not available',
        'policy_documents' : 'policy.',
        # 'clinical_trials' : 'not available',
        # 'datasets' : 'not available'
        'researchers' : 'ur.',
        'organizations' : 'grid.',
    },
    'lang_all': [
        'search',
        'return',
        'for',
        'where',
        'in',
        'limit',
        'skip',
        'aggregate',
        '=',  # filter operators https://docs.dimensions.ai/dsl/language.html#simple-filters
        '!=',
        '>',
        '<',
        '>=',
        '<=',
        '~',
        'is empty',
        'is not empty',
        "count", # https://docs.dimensions.ai/dsl/language.html#filter-functions
        'sort by',
        'asc',
        'desc',
        "AND", # boolean operators https://docs.dimensions.ai/dsl/language.html#id6
        "OR", 
        "NOT",
        "&&",
        "!",
        "||",
        "+",
        "-",
    ],
    'lang_after_search' : ['in', 'where', 'for', 'return'],
    'lang_after_filter' : ['and', 'or', 'not', 'return', ],
    'lang_after_for_text' : ['and', 'or', 'not', 'return', 'where' ],
    'lang_after_return' : ['sort by', 'aggregate', 'limit',],
    'lang_after_sort_by' : ['asc', 'desc', 'limit', ],
    'lang_after_limit' : ['skip' ],
    'lang_filter_operators' : ['=', '!=', '>', '<', '>=', '<=', '~', 'is empty', 'is not empty'],
    'lang_text_operators' : ['AND', 'OR', 'NOT', '&&', '!', '||', '+', '-', '?', '*', '~'],
}


#
# GRAMMAR_DICT is a dictionary rendering of the DSL grammar JSON
# which can be obtained with the query `describe schema`
#
# last updated: v 1.19 2019-09-05
#
# how to create:
#
# In [1]: import dimcli
# In [2]: dimcli.login()
# In [3]: dsl = dimcli.Dsl()
# In [4]: dsl.query("describe schema").json
#
# then save to a py file, reformat and save the results in GRAMMAR_DICT symbol
#
#
GRAMMAR_DICT = {
    'sources': {
        'publications': {
            'fields': {
                'doi': {
                    'type': 'string',
                    'description': 'Digital object identifier.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'research_org_names': {
                    'type':
                    'string',
                    'description':
                    'Names of organizations authors are affiliated to.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'open_access_categories': {
                    'type':
                    'open_access',
                    'description':
                    'Open Access categories for publications. See below for more examples.',
                    'long_description':
                    'Open Access category data for publications values:\n\n        * `oa_all`: Article is freely available\n        * `gold_pure`: Version Of Record (VOR) is free under an open licence from a full OA journal\n        * `gold_hybrid`: Version Of Record (VOR) is free under an open licence in a paid-access journal\n        * `gold_bronze`: Freely available on publisher page, but without an open licence\n        * `green_pub`: Free copy of published version in an OA repository\n        * `green_acc`: Free copy of accepted version in an OA repository\n        * `green_sub`: Free copy of submitted version, or where version is unknown, in an OA repository\n        * `closed`: No freely available copy has been identified',
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_uoa': {
                    'type':
                    'categories',
                    'description':
                    '`Units of Assessment <https://app.dimensions.ai/browse/publication/uoa>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'resulting_publication_doi': {
                    'type':
                    'string',
                    'description':
                    'For preprints, the DOIs of the resulting full publications.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'altmetric': {
                    'type': 'float',
                    'description': 'Altmetric attention score.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'category_for': {
                    'type':
                    'categories',
                    'description':
                    '`ANZSRC Fields of Research classification <https://app.dimensions.ai/browse/publication/for>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'research_org_state_names': {
                    'type':
                    'string',
                    'description':
                    'State name of the organisations authors are affiliated to, as a string.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'category_hrcs_hc': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS - Health Categories <https://app.dimensions.ai/browse/publication/hrcs_hc>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hra': {
                    'type':
                    'categories',
                    'description':
                    '`Health Research Areas <https://app.dimensions.ai/browse/publication/health_research_areas>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_rcdc': {
                    'type':
                    'categories',
                    'description':
                    '`Research, Condition, and Disease Categorization <https://app.dimensions.ai/browse/publication/rcdc>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'linkout': {
                    'type': 'string',
                    'description': 'Original URL for a publication full text.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'publisher': {
                    'type': 'string',
                    'description': 'Name of the publisher as a string.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'category_sdg': {
                    'type': 'categories',
                    'description': 'SDG - Sustainable Development Goals',
                    'long_description': None,
                    'is_entity': True,
                    'is_filter': True,
                    'is_facet': True
                },
                'issue': {
                    'type': 'string',
                    'description': 'The issue number of a publication.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'pmcid': {
                    'type': 'string',
                    'description': 'PubMed Central ID.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'mesh_terms': {
                    'type':
                    'string',
                    'description':
                    'Medical Subject Heading terms as used in PubMed.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'funder_countries': {
                    'type':
                    'countries',
                    'description':
                    'The country of the organisations funding this publication.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'research_org_state_codes': {
                    'type':
                    'states',
                    'description':
                    'State of the organisations authors are affiliated to, expressed as GeoNames codes (ISO\u200c-3166-2).',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'pages': {
                    'type':
                    'string',
                    'description':
                    'The pages of the publication, as they would appear in a citation record.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'book_series_title': {
                    'type':
                    'string',
                    'description':
                    'The title of the book series book, belong to.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    False,
                    'is_facet':
                    False
                },
                'proceedings_title': {
                    'type':
                    'string',
                    'description':
                    'Title of the conference proceedings volume associated to a publication.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    False,
                    'is_facet':
                    False
                },
                'category_hrcs_rac': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS – Research Activity Codes <https://app.dimensions.ai/browse/publication/hrcs_rac>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'research_org_cities': {
                    'type':
                    'cities',
                    'description':
                    'City of the organisations authors are affiliated to, expressed as GeoNames ID and name.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'research_org_country_names': {
                    'type':
                    'string',
                    'description':
                    'Country name of the organisations authors are affiliated to, as a string.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'date': {
                    'type':
                    'date',
                    'description':
                    'The publication date of a document, eg "2018-01-01" (note: dates can sometimes be incomplete and include only the month or the year).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'book_title': {
                    'type':
                    'string',
                    'description':
                    'The title of the book a chapter belongs to (note: this field is available only for chapters).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    False,
                    'is_facet':
                    False
                },
                'funders': {
                    'type':
                    'organizations',
                    'description':
                    'The GRID organisation funding this publication.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'date_inserted': {
                    'type':
                    'date',
                    'description':
                    "Date when the record was inserted into Dimensions (note: this field does not support exact match on the data, only range filters e.g. `<=` or `>=').",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'research_org_countries': {
                    'type':
                    'countries',
                    'description':
                    'Country of the organisations authors are affiliated to, identified using GeoNames codes (note: this field supports :ref:`filter-functions`: ``count``).',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'supporting_grant_ids': {
                    'type':
                    'string',
                    'description':
                    'Grants supporting a publication, returned as a list of dimensions grants IDs  (see also: :ref:`publications_model` section).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'altmetric_id': {
                    'type': 'integer',
                    'description': 'AltMetric Publication ID',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'book_doi': {
                    'type':
                    'string',
                    'description':
                    'The DOI of the book a chapter belongs to (note: this field is available only for chapters).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'volume': {
                    'type': 'string',
                    'description': 'Publication volume.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'issn': {
                    'type': 'string',
                    'description': 'International Standard Serial Number',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'journal': {
                    'type': 'journals',
                    'description': 'The journal a publication belongs to.',
                    'long_description': None,
                    'is_entity': True,
                    'is_filter': True,
                    'is_facet': True
                },
                'concepts': {
                    'type': 'json',
                    'description': None,
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'authors': {
                    'type':
                    'json',
                    'description':
                    'Ordered list of authors names and their affiliations, as they appear in the original publication. The list can include researcher and organization identifiers, when available (note: in order to search for disambiguated authors, use the `in researchers` syntax).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'research_orgs': {
                    'type':
                    'organizations',
                    'description':
                    'GRID organisations associated to a publication. Identifiers are automatically extracted from author affiliations text, so they can be missing in some cases (note: this field supports :ref:`filter-functions`: ``count``).',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'relative_citation_ratio': {
                    'type':
                    'float',
                    'description':
                    'Relative citation performance of an article when compared to others in its area of research (note: does not support emptiness filters).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'researchers': {
                    'type':
                    'researchers',
                    'description':
                    "Researcher IDs matched to the publication's authors list. (note: this returns only the disambiguated authors of a publication; in order to get the full authors list, the field `authors` should be used). This field supports :ref:`filter-functions`: ``count``.",
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'journal_lists': {
                    'type':
                    'string',
                    'description':
                    "Independent grouping of journals outside of Dimensions, e.g. 'ERA 2015' or 'Norwegian register level 1'.",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'field_citation_ratio': {
                    'type':
                    'float',
                    'description':
                    'Relative citation performance of article when compared to similarly aged articles in its area of research (note: does not support emptiness filters).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'concepts_scores': {
                    'type': 'json',
                    'description': None,
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'reference_ids': {
                    'type':
                    'string',
                    'description':
                    'Dimensions publication ID for publications in the references list, i.e. outgoing citations (see also: :ref:`publications_model` section).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'category_icrp_cso': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Common Scientific Outline <https://app.dimensions.ai/browse/publication/cso>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'id': {
                    'type': 'string',
                    'description': 'Dimensions publication ID.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'type': {
                    'type':
                    'string',
                    'description':
                    'Publication type (one of: article, chapter, proceeding, monograph, preprint or book).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'pmid': {
                    'type': 'string',
                    'description': 'PubMed ID.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'recent_citations': {
                    'type':
                    'integer',
                    'description':
                    'Number of citations received in the last two years. Does not support emptiness filters',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'title': {
                    'type': 'string',
                    'description': 'Title of a publication.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'times_cited': {
                    'type':
                    'integer',
                    'description':
                    'Number of citations (note: does not support emptiness filters).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_icrp_ct': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Cancer Types <https://app.dimensions.ai/browse/publication/cancer_types>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'year': {
                    'type':
                    'integer',
                    'description':
                    'The year of publication (note: when the `date` field is available, this is equal to the year part of the full date).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_bra': {
                    'type':
                    'categories',
                    'description':
                    '`Broad Research Areas <https://app.dimensions.ai/browse/publication/broad_research_areas>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                }
            },
            'fieldsets': ['all', 'basics', 'extras', 'book', 'categories'],
            'metrics': {
                'count': {
                    'name': 'count',
                    'description': 'Total count'
                },
                'altmetric_median': {
                    'name': 'altmetric_median',
                    'description': 'Median Altmetric attention score'
                },
                'altmetric_avg': {
                    'name': 'altmetric_avg',
                    'description': 'Altmetric attention score mean'
                },
                'citations_total': {
                    'name': 'citations_total',
                    'description': 'Aggregated number of citations'
                },
                'citations_avg': {
                    'name': 'citations_avg',
                    'description': 'Arithmetic mean of citations'
                },
                'citations_median': {
                    'name': 'citations_median',
                    'description': 'Median of citations'
                },
                'recent_citations_total': {
                    'name':
                    'recent_citations_total',
                    'description':
                    'For a given article, in a given year, the number of citations accrued in the last two year period. Single value stored per document, year window rolls over in July.'
                },
                'rcr_avg': {
                    'name':
                    'rcr_avg',
                    'description':
                    'Arithmetic mean of `relative_citation_ratio` field.'
                },
                'fcr_gavg': {
                    'name':
                    'fcr_gavg',
                    'description':
                    'Geometric mean of `field_citation_ratio` field (note: This field cannot be used for sorting results).'
                }
            },
            'search_fields': [
                'noun_phrases', 'authors', 'full_data_exact', 'full_data',
                'title_abstract_only', 'title_only', 'concepts'
            ]
        },
        'grants': {
            'fields': {
                'category_uoa': {
                    'type':
                    'categories',
                    'description':
                    '`Units of Assessment <https://app.dimensions.ai/browse/publication/uoa>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'funding_aud': {
                    'type': 'float',
                    'description': 'Funding amount awarded in AUD.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'category_for': {
                    'type':
                    'categories',
                    'description':
                    '`ANZSRC Fields of Research classification <https://app.dimensions.ai/browse/publication/for>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hrcs_hc': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS - Health Categories <https://app.dimensions.ai/browse/publication/hrcs_hc>`_ .',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hra': {
                    'type':
                    'categories',
                    'description':
                    '`Health Research Areas <https://app.dimensions.ai/browse/publication/health_research_areas?redirect_path=/discover/publication>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'linkout': {
                    'type': 'string',
                    'description': 'Original URL for the grant.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'category_rcdc': {
                    'type':
                    'categories',
                    'description':
                    '`Research, Condition, and Disease Categorization <https://app.dimensions.ai/browse/publication/rcdc>`_ .',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'end_date': {
                    'type': 'date',
                    'description': 'Date when the grant ends.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'investigator_details': {
                    'type':
                    'json',
                    'description':
                    "Additional details about investigators, including affiliations and roles e.g. 'PI' or 'Co-PI' (note: if the investigator has a Dimensions researcher ID, that is returned as well).",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'language': {
                    'type':
                    'string',
                    'description':
                    'Grant original language, as ISO 639-1 language codes.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'funder_countries': {
                    'type':
                    'countries',
                    'description':
                    'The country linked to the organisation funding the grant, expressed as GeoNames codes.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'funding_gbp': {
                    'type': 'float',
                    'description': 'Funding amount awarded in GBP.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'research_org_state_codes': {
                    'type':
                    'states',
                    'description':
                    'State of the organisations receiving the grant, expressed as GeoNames codes (ISO\u200c-3166-2).',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hrcs_rac': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS – Research Activity Codes <https://app.dimensions.ai/browse/publication/hrcs_rac>`_ .',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'research_org_cities': {
                    'type':
                    'cities',
                    'description':
                    'City of the research organisations receiving the grant, expressed as GeoNames id and name.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'funding_jpy': {
                    'type': 'float',
                    'description': 'Funding amount awarded in JPY.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'funding_nzd': {
                    'type': 'float',
                    'description': 'Funding amount awarded in NZD.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'start_year': {
                    'type': 'integer',
                    'description': 'Year when the grant starts.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'funders': {
                    'type':
                    'organizations',
                    'description':
                    'The organisation funding the grant. This is normally a GRID organisation, but in very few cases a Dimensions funder ID is used.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'funding_currency': {
                    'type': 'string',
                    'description': 'Original funding currency.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'date_inserted': {
                    'type':
                    'date',
                    'description':
                    'Date when the record was inserted into Dimensions (note: this field does not support exact match on the data, only range filters e.g. `<=` or `>=`).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'grant_number': {
                    'type':
                    'string',
                    'description':
                    'Grant identifier, as provided by the source (e.g., funder, aggregator) the grant was derived from.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'research_org_countries': {
                    'type':
                    'countries',
                    'description':
                    'Country of the research organisations receiving the grant, expressed as GeoNames code and name.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'active_year': {
                    'type': 'integer',
                    'description': 'List of active years for a grant.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'funding_org_city': {
                    'type': 'string',
                    'description': 'City name for funding organisation.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'funding_chf': {
                    'type': 'float',
                    'description': 'Funding amount awarded in CHF.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'funding_org_name': {
                    'type': 'string',
                    'description': 'Name of funding organisation.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'language_title': {
                    'type':
                    'string',
                    'description':
                    'ISO 639-1 language code for the original grant title.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'foa_number': {
                    'type':
                    'string',
                    'description':
                    'The funding opportunity announcement (FOA) number, where available e.g. for grants from the US National Institute of Health (NIH) or from the National Science Foundation (NSF).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'funding_eur': {
                    'type': 'float',
                    'description': 'Funding amount awarded in EUR.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'original_title': {
                    'type': 'string',
                    'description':
                    'Title of the grant in its original language.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'concepts': {
                    'type':
                    'string',
                    'description':
                    'Concepts describing the main topics of a grant (note: automatically derived from the grant text using machine learning).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    False,
                    'is_facet':
                    False
                },
                'funding_org_acronym': {
                    'type': 'string',
                    'description': 'Acronym for funding organisation.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'research_orgs': {
                    'type':
                    'organizations',
                    'description':
                    'GRID organisations receiving the grant (note: identifiers are automatically extracted from the source text and can be missing in some cases).',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'researchers': {
                    'type':
                    'researchers',
                    'description':
                    'Dimensions researchers IDs associated to the grant.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'abstract': {
                    'type': 'string',
                    'description':
                    'Abstract or summary from a grant proposal.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'category_icrp_cso': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Common Scientific Outline <https://app.dimensions.ai/browse/publication/cso>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'id': {
                    'type': 'string',
                    'description': 'Dimensions grant ID.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'funding_usd': {
                    'type': 'float',
                    'description': 'Funding amount awarded in USD.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'title': {
                    'type':
                    'string',
                    'description':
                    'Title of the grant in English (if the grant language is not English, this field contains a translation of the title).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    False,
                    'is_facet':
                    False
                },
                'start_date': {
                    'type':
                    'date',
                    'description':
                    "Date when the grant starts, in the format 'YYYY-MM-DD'.",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'funding_cad': {
                    'type': 'float',
                    'description': 'Funding amount awarded in CAD.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'category_icrp_ct': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Cancer Types <https://app.dimensions.ai/browse/publication/cancer_types>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_bra': {
                    'type':
                    'categories',
                    'description':
                    '`Broad Research Areas <https://app.dimensions.ai/browse/publication/broad_research_areas?redirect_path=/discover/publication>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                }
            },
            'fieldsets': ['all', 'basics', 'extras', 'categories'],
            'metrics': {
                'count': {
                    'name': 'count',
                    'description': 'Total count'
                },
                'funding': {
                    'name': 'funding',
                    'description': 'Total funding amount, in USD.'
                }
            },
            'search_fields': [
                'noun_phrases', 'full_data', 'investigators',
                'title_abstract_only', 'title_only', 'concepts'
            ]
        },
        'patents': {
            'fields': {
                'associated_grant_ids': {
                    'type':
                    'string',
                    'description':
                    'Dimensions IDs of the grants associated to the patent (see also: :ref:`patents_model` section).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'assignee_cities': {
                    'type':
                    'cities',
                    'description':
                    'City of the assignees of the patent, expressed as GeoNames ID and name (note: this value is extracted independently from GRID).',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'legal_status': {
                    'type':
                    'string',
                    'description':
                    "The legal status of the patent, e.g. 'Granted', 'Active', 'Abandoned' etc..",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'jurisdiction': {
                    'type':
                    'string',
                    'description':
                    "The jurisdiction where the patent was granted, e.g. 'US', 'DE', 'EP'...",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'assignee_countries': {
                    'type':
                    'countries',
                    'description':
                    'Country of the assignees of the patent, expressed as GeoNames code and name (note: this value is extracted independently from GRID).',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_for': {
                    'type':
                    'categories',
                    'description':
                    '`ANZSRC Fields of Research classification <https://app.dimensions.ai/browse/publication/for>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'publication_year': {
                    'type': 'integer',
                    'description': 'Year of publication of a patent.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'category_hrcs_hc': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS - Health Categories <https://app.dimensions.ai/browse/publication/hrcs_hc>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hra': {
                    'type':
                    'categories',
                    'description':
                    '`Health Research Areas <https://app.dimensions.ai/browse/publication/health_research_areas?redirect_path=/discover/publication>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_rcdc': {
                    'type':
                    'categories',
                    'description':
                    '`Research, Condition, and Disease Categorization <https://app.dimensions.ai/browse/publication/rcdc>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'granted_year': {
                    'type':
                    'integer',
                    'description':
                    'The year on which the official body grants the patent.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'assignees': {
                    'type':
                    'organizations',
                    'description':
                    'Disambiguated GRID organisations who own or have owned the rights of a patent (note: this is a combination of `current_assignees` and `original_assignees` fields).',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hrcs_rac': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS – Research Activity Codes <https://app.dimensions.ai/browse/publication/hrcs_rac>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'expiration_date': {
                    'type': 'date',
                    'description': 'Date when the patent expires.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'date': {
                    'type': 'date',
                    'description': 'Date when the patent was filed.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'funders': {
                    'type': 'organizations',
                    'description': 'GRID organisations funding the patent.',
                    'long_description': None,
                    'is_entity': True,
                    'is_filter': True,
                    'is_facet': True
                },
                'date_inserted': {
                    'type':
                    'date',
                    'description':
                    'Date when the record was inserted into Dimensions (note: this field does not support exact match on the data, only range filters e.g. `<=` or `>=`).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'cpc': {
                    'type':
                    'string',
                    'description':
                    '`Cooperative Patent Classification Categorization <https://www.epo.org/searching-for-patents/helpful-resources/first-time-here/classification/cpc.html>`_.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'filing_status': {
                    'type':
                    'string',
                    'description':
                    "Filing Status of the patent e.g. 'Application' or 'Grant'.",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'additional_filters': {
                    'type':
                    'string',
                    'description':
                    "Additional filters describing the patents, e.g. whether it's about a 'Research Organisation', or it is part of the 'Orange Book'.",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'priority_year': {
                    'type':
                    'integer',
                    'description':
                    'The filing year of the earliest application of which priority is claimed.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'publication_ids': {
                    'type':
                    'string',
                    'description':
                    'Dimensions IDs of the publications related to this patent (see also: :ref:`patents_model` section).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'assignee_state_codes': {
                    'type':
                    'states',
                    'description':
                    'State of the assignee, expressed using GeoNames (ISO\u200c-3166-2) codes (note: this value is extracted independently from GRID).',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'original_assignees': {
                    'type':
                    'organizations',
                    'description':
                    'Disambiguated GRID organisations that first owned the patent.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'assignee_names': {
                    'type':
                    'string',
                    'description':
                    'Name of assignees of the patent, as they appear in the original document.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'current_assignee_names': {
                    'type':
                    'string',
                    'description':
                    'Names of the organisations currently holding the patent, as they appear in the original document.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'granted_date': {
                    'type':
                    'date',
                    'description':
                    'The date on which the official body grants the patent.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'original_assignee_names': {
                    'type':
                    'string',
                    'description':
                    'Name of the organisations that first owned the patent, as they appear in the original document.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'publication_date': {
                    'type': 'date',
                    'description': 'Date of publication of a patent.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'ipcr': {
                    'type':
                    'string',
                    'description':
                    '`International Patent Classification Reform Categorization <https://www.wipo.int/classifications/ipc/en/faq/>`_.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'inventor_names': {
                    'type': 'string',
                    'description':
                    'Names of the people who invented the patent.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'priority_date': {
                    'type':
                    'date',
                    'description':
                    'The earliest filing date in a family of patent applications.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'current_assignees': {
                    'type':
                    'organizations',
                    'description':
                    'Disambiguated GRID organisations currenlty owning the patent.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'researchers': {
                    'type':
                    'researchers',
                    'description':
                    "Researcher IDs matched to the patent's inventors list. (note: this returns only the disambiguated inventors of a patent; in order to get the full list of inventors, the field `inventors` should be used).",
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'abstract': {
                    'type': 'string',
                    'description': 'Abstract or description of the patent.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'reference_ids': {
                    'type':
                    'string',
                    'description':
                    'Dimensions IDs of the patents which are cited by this patent (see also: :ref:`patents_model` section).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'category_icrp_cso': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Common Scientific Outline <https://app.dimensions.ai/browse/publication/cso>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'id': {
                    'type': 'string',
                    'description': 'Dimensions patent ID',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'title': {
                    'type': 'string',
                    'description': 'The title of the patent.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'times_cited': {
                    'type':
                    'integer',
                    'description':
                    'The number of times the patent has been cited by other patents.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_icrp_ct': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Cancer Types <https://app.dimensions.ai/browse/publication/cancer_types>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'year': {
                    'type': 'integer',
                    'description': 'The year the patent was filed.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'category_bra': {
                    'type':
                    'categories',
                    'description':
                    '`Broad Research Areas <https://app.dimensions.ai/browse/publication/broad_research_areas?redirect_path=/discover/publication>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'cited_by_ids': {
                    'type':
                    'string',
                    'description':
                    'Dimensions IDs of the patents that cite this patent (see also: :ref:`patents_model` section).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                }
            },
            'fieldsets': ['all', 'basics', 'extras', 'categories'],
            'metrics': {
                'count': {
                    'name': 'count',
                    'description': 'Total count'
                }
            },
            'search_fields':
            ['inventors', 'full_data', 'title_only', 'title_abstract_only']
        },
        'clinical_trials': {
            'fields': {
                'acronym': {
                    'type': 'string',
                    'description': 'Acronym of the clinical trial.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'associated_grant_ids': {
                    'type':
                    'string',
                    'description':
                    'Dimensions IDs of the grants associated to the clinical trial (see also: :ref:`clinical_trials_model` section).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'phase': {
                    'type': 'string',
                    'description': 'Phase of the clinical trial, as a string.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'category_for': {
                    'type':
                    'categories',
                    'description':
                    '`ANZSRC Fields of Research classification <https://app.dimensions.ai/browse/publication/for>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'conditions': {
                    'type':
                    'string',
                    'description':
                    "List of medical conditions names, e.g. 'Breast cancer' or 'Obesity'.",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'category_hrcs_hc': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS - Health Categories <https://app.dimensions.ai/browse/publication/hrcs_hc>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hra': {
                    'type':
                    'categories',
                    'description':
                    '`Health Research Areas <https://app.dimensions.ai/browse/publication/health_research_areas?redirect_path=/discover/publication>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'linkout': {
                    'type': 'string',
                    'description': 'Original URL for the clinical trial.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'category_rcdc': {
                    'type':
                    'categories',
                    'description':
                    '`Research, Condition, and Disease Categorization <https://app.dimensions.ai/browse/publication/rcdc>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'investigator_details': {
                    'type':
                    'json',
                    'description':
                    'Additional details about investigators, including affiliations and roles.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'funder_countries': {
                    'type': 'countries',
                    'description':
                    'The country group the funding organisations.',
                    'long_description': None,
                    'is_entity': True,
                    'is_filter': True,
                    'is_facet': True
                },
                'category_hrcs_rac': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS – Research Activity Codes <https://app.dimensions.ai/browse/publication/hrcs_rac>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'gender': {
                    'type':
                    'string',
                    'description':
                    "The gender of the clinical trial subjects e.g. 'Male', 'Female' or 'All'.",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'funders': {
                    'type':
                    'organizations',
                    'description':
                    'GRID funding organisations that are involved with the clinical trial.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'date': {
                    'type': 'date',
                    'description': 'Start date of a clinical trial.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'date_inserted': {
                    'type':
                    'date',
                    'description':
                    "Date when the record was inserted into Dimensions (note: this field does not support exact match on the data, only range filters e.g. `<=` or `>=').",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'publication_ids': {
                    'type':
                    'string',
                    'description':
                    'Dimensions IDs of the publications related to this clinical trial (see also: :ref:`clinical_trials_model` section).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'brief_title': {
                    'type': 'string',
                    'description': 'Brief title of the clinical trial.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'interventions': {
                    'type':
                    'json',
                    'description':
                    "Structured JSON object containing information about the clinical trial's interventions according to the research plan or protocol created by the investigators.",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'active_years': {
                    'type': 'integer',
                    'description':
                    'List of active years for a clinical trial.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'research_orgs': {
                    'type':
                    'organizations',
                    'description':
                    'GRID organizations involved, e.g. as sponsors or collaborators.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'researchers': {
                    'type':
                    'researchers',
                    'description':
                    'Dimensions researchers IDs associated to the clinical trial.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'abstract': {
                    'type':
                    'string',
                    'description':
                    'Abstract or description of the clinical trial.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    False,
                    'is_facet':
                    False
                },
                'category_icrp_cso': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Common Scientific Outline <https://app.dimensions.ai/browse/publication/cso>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'id': {
                    'type': 'string',
                    'description': 'Dimensions clinical trial ID',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'title': {
                    'type': 'string',
                    'description': 'The title of the clinical trial.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'registry': {
                    'type':
                    'string',
                    'description':
                    "The platform where the clinical trial has been registered, e.g. 'ClinicalTrials.gov' or 'EU-CTR'.",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'category_icrp_ct': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Cancer Types <https://app.dimensions.ai/browse/publication/cancer_types>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_bra': {
                    'type':
                    'categories',
                    'description':
                    '`Broad Research Areas <https://app.dimensions.ai/browse/publication/broad_research_areas?redirect_path=/discover/publication>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                }
            },
            'fieldsets': ['all', 'basics', 'extras', 'categories'],
            'metrics': {
                'count': {
                    'name': 'count',
                    'description': 'Total count'
                }
            },
            'search_fields': [
                'investigators', 'full_data', 'title_only',
                'title_abstract_only'
            ]
        },
        'policy_documents': {
            'fields': {
                'category_for': {
                    'type':
                    'categories',
                    'description':
                    '`ANZSRC Fields of Research classification <https://app.dimensions.ai/browse/publication/for>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hrcs_hc': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS - Health Categories <https://app.dimensions.ai/browse/publication/hrcs_hc>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hra': {
                    'type':
                    'categories',
                    'description':
                    '`Health Research Areas <https://app.dimensions.ai/browse/publication/health_research_areas?redirect_path=/discover/publication>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_rcdc': {
                    'type':
                    'categories',
                    'description':
                    '`Research, Condition, and Disease Categorization <https://app.dimensions.ai/browse/publication/rcdc>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'linkout': {
                    'type': 'string',
                    'description': 'Original URL for the policy document.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'publisher_org_state': {
                    'type':
                    'states',
                    'description':
                    'State of the organization publishing the policy document.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'publisher_org_country': {
                    'type':
                    'countries',
                    'description':
                    'Country of the organization publishing the policy document.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'publisher_org_city': {
                    'type':
                    'cities',
                    'description':
                    'City of the organization publishing the policy document.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hrcs_rac': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS – Research Activity Codes <https://app.dimensions.ai/browse/publication/hrcs_rac>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'date_inserted': {
                    'type':
                    'date',
                    'description':
                    'Date when the record was inserted into Dimensions (note: this field does not support exact match on the data, only range filters e.g. `<=` or `>=`).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'publication_ids': {
                    'type':
                    'string',
                    'description':
                    'Dimensions IDs of the publications related to this policy document (see also: :ref:`policy_documents_model` section).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'publisher_org': {
                    'type':
                    'organizations',
                    'description':
                    'GRID organization publishing the policy document.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_icrp_cso': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Common Scientific Outline <https://app.dimensions.ai/browse/publication/cso>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'id': {
                    'type': 'string',
                    'description': 'Dimensions policy document ID',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'title': {
                    'type': 'string',
                    'description': 'Title of the policy document.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'category_icrp_ct': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Cancer Types <https://app.dimensions.ai/browse/publication/cancer_types>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'year': {
                    'type': 'integer',
                    'description':
                    'Year of publication of the policy document.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'category_bra': {
                    'type':
                    'categories',
                    'description':
                    '`Broad Research Areas <https://app.dimensions.ai/browse/publication/broad_research_areas?redirect_path=/discover/publication>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                }
            },
            'fieldsets': ['all', 'basics', 'categories'],
            'metrics': {
                'count': {
                    'name': 'count',
                    'description': 'Total count'
                }
            },
            'search_fields': ['full_data', 'title_only']
        },
        'researchers': {
            'fields': {
                'research_orgs': {
                    'type':
                    'organizations',
                    'description':
                    'All research organizations linked to the researcher.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'total_grants': {
                    'type': 'integer',
                    'description': 'Total grants count.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'nih_ppid': {
                    'type':
                    'string',
                    'description':
                    'The PI Profile ID (i.e., ppid) is a Researcher ID from the US National Institute of Health (NIH).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'first_publication_year': {
                    'type': 'integer',
                    'description': None,
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'first_name': {
                    'type': 'string',
                    'description': 'First Name.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'last_name': {
                    'type': 'string',
                    'description': 'Last Name.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'id': {
                    'type': 'string',
                    'description': 'Dimensions researcher ID.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'last_grant_year': {
                    'type':
                    'integer',
                    'description':
                    'Last year the researcher was awarded a grant.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'redirect': {
                    'type':
                    'string',
                    'description':
                    'Indicates status of a researcher ID marked as obsolete. Empty means that the researcher ID was deleted. Otherwise ID provided means that is the new ID into which the obsolete one was redirected. If multiple values are available, it means that the original researcher ID was split into multiple IDs.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'last_publication_year': {
                    'type': 'integer',
                    'description': None,
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'current_research_org': {
                    'type':
                    'organizations',
                    'description':
                    'The most recent research organization linked to the researcher.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'obsolete': {
                    'type':
                    'integer',
                    'description':
                    'Indicates researcher ID status. 0 means that the researcher ID is still active, 1 means that the researcher ID is no longer valid. See the `redirect` field for further information on invalid researcher IDs.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'total_publications': {
                    'type': 'integer',
                    'description': 'Total publications count.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'orcid_id': {
                    'type': 'string',
                    'description': '`ORCID <https://orcid.org/>`_ ID.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'first_grant_year': {
                    'type':
                    'integer',
                    'description':
                    'First year the researcher was awarded a grant.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                }
            },
            'fieldsets': ['all', 'basics', 'extras'],
            'metrics': {
                'count': {
                    'name': 'count',
                    'description': 'Total count'
                }
            },
            'search_fields': []
        },
        'organizations': {
            'fields': {
                'acronym': {
                    'type':
                    'string',
                    'description':
                    'GRID acronym of the organization. E.g., "UT" for `grid.26999.3d <https://grid.ac/institutes/grid.26999.3d>`_',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'wikidata_ids': {
                    'type': 'string',
                    'description': 'WikiData IDs for this organization',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'types': {
                    'type':
                    'string',
                    'description':
                    'Type of an organization. Available types include: ``Company``, ``Education``, ``Healthcare``, ``Nonprofit``, ``Facility``, ``Other``, ``Government``, ``Archive``, ``Education,Company``, ``Education,Facility``, ``Education,Healthcare``, ``Education,Other``, ``Archive,Nonprofit``. Furhter explanation is on the `GRID <https://www.grid.ac/pages/policies>`_ website.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'established': {
                    'type': 'integer',
                    'description':
                    'Year when the organization was estabilished',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'nuts_level2_code': {
                    'type':
                    'string',
                    'description':
                    'Level 2 code for this organization, based on `Nomenclature of Territorial Units for Statistics <https://en.wikipedia.org/wiki/Nomenclature_of_Territorial_Units_for_Statistics>`_ (NUTS) codes of the European Union.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'nuts_level1_name': {
                    'type':
                    'string',
                    'description':
                    'Level 1 name for this organization, based on `Nomenclature of Territorial Units for Statistics <https://en.wikipedia.org/wiki/Nomenclature_of_Territorial_Units_for_Statistics>`_ (NUTS) codes of the European Union.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'linkout': {
                    'type': 'string',
                    'description': None,
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'latitude': {
                    'type': 'float',
                    'description': None,
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'name': {
                    'type':
                    'string',
                    'description':
                    'GRID name of the organization. E.g., "University of Tokyo" for `grid.26999.3d <https://grid.ac/institutes/grid.26999.3d>`_',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'ucas_ids': {
                    'type': 'string',
                    'description': 'UCAS IDs for this organization',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'nuts_level2_name': {
                    'type':
                    'string',
                    'description':
                    'Level 2 name for this organization, based on `Nomenclature of Territorial Units for Statistics <https://en.wikipedia.org/wiki/Nomenclature_of_Territorial_Units_for_Statistics>`_ (NUTS) codes of the European Union.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'organization_child_ids': {
                    'type': 'string',
                    'description': 'Child organization IDs',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'cnrs_ids': {
                    'type': 'string',
                    'description': 'CNRS IDs for this organization',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'country_name': {
                    'type':
                    'string',
                    'description':
                    'GRID name of the organization country. E.g., "Japan" for `grid.26999.3d <https://grid.ac/institutes/grid.26999.3d>`_',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'ukprn_ids': {
                    'type': 'string',
                    'description': 'UKPRN IDs for this organization',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'wikipedia_url': {
                    'type': 'string',
                    'description': 'Wikipedia URL',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'city_name': {
                    'type':
                    'string',
                    'description':
                    'GRID name of the organization country. E.g., "Bethesda" for `grid.419635.c <https://grid.ac/institutes/grid.419635.c>`_',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'hesa_ids': {
                    'type': 'string',
                    'description': 'HESA IDs for this organization',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'longitude': {
                    'type': 'float',
                    'description': None,
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'organization_related_ids': {
                    'type': 'string',
                    'description': 'Related organization IDs',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'orgref_ids': {
                    'type': 'string',
                    'description': 'OrgRef IDs for this organization',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'organization_parent_ids': {
                    'type': 'string',
                    'description': 'Parent organization IDs',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'state_name': {
                    'type':
                    'string',
                    'description':
                    'GRID name of the organization country. E.g., "Maryland" for `grid.419635.c <https://grid.ac/institutes/grid.419635.c>`_',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'id': {
                    'type':
                    'string',
                    'description':
                    'GRID ID of the organization. E.g., "grid.26999.3d".',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'nuts_level1_code': {
                    'type':
                    'string',
                    'description':
                    'Level 1 code for this organization, based on `Nomenclature of Territorial Units for Statistics <https://en.wikipedia.org/wiki/Nomenclature_of_Territorial_Units_for_Statistics>`_ (NUTS) codes of the European Union.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'isni_ids': {
                    'type': 'string',
                    'description': 'ISNI IDs for this organization',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'external_ids_fundref': {
                    'type': 'string',
                    'description': 'Fundref IDs for this organization',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                }
            },
            'fieldsets': ['all', 'basics', 'nuts'],
            'metrics': {
                'count': {
                    'name': 'count',
                    'description': 'Total count'
                }
            },
            'search_fields': ['full_data']
        },
        'datasets': {
            'fields': {
                'doi': {
                    'type': 'string',
                    'description': 'Dataset DOI.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'associated_grant_ids': {
                    'type':
                    'string',
                    'description':
                    'The Dimensions IDs of the grants linked to the publication the dataset is associated with.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'date_embargo': {
                    'type': 'date',
                    'description': 'The embargo date of the dataset.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'category_for': {
                    'type':
                    'categories',
                    'description':
                    '`ANZSRC Fields of Research classification <https://app.dimensions.ai/browse/publication/for>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hrcs_hc': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS - Health Categories <https://app.dimensions.ai/browse/publication/hrcs_hc>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_hra': {
                    'type':
                    'categories',
                    'description':
                    '`Health Research Areas <https://app.dimensions.ai/browse/publication/health_research_areas>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_rcdc': {
                    'type':
                    'categories',
                    'description':
                    '`Research, Condition, and Disease Categorization <https://app.dimensions.ai/browse/publication/rcdc>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'funder_countries': {
                    'type':
                    'countries',
                    'description':
                    'The country linked to the organisation funding the grant, expressed as GeoNames codes.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'keywords': {
                    'type':
                    'string',
                    'description':
                    'Keywords used to describe the dataset (from authors).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'license': {
                    'type':
                    'json',
                    'description':
                    'The dataset licence, as a structured JSON containing the license name, URL, and value.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'category_hrcs_rac': {
                    'type':
                    'categories',
                    'description':
                    '`HRCS – Research Activity Codes <https://app.dimensions.ai/browse/publication/hrcs_rac>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'research_org_cities': {
                    'type':
                    'cities',
                    'description':
                    'City of the organisations the publication authors are affiliated to, expressed as GeoNames ID and name.',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'funders': {
                    'type': 'organizations',
                    'description':
                    'The GRID organisations funding the dataset.',
                    'long_description': None,
                    'is_entity': True,
                    'is_filter': True,
                    'is_facet': True
                },
                'date': {
                    'type':
                    'date',
                    'description':
                    'The publication date of the dataset, eg "2018-01-01".',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'date_inserted': {
                    'type':
                    'date',
                    'description':
                    "Date when the record was inserted into Dimensions (note: this field does not support exact match on the data, only range filters e.g. `<=` or `>=').",
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'research_org_countries': {
                    'type':
                    'countries',
                    'description':
                    'Country of the organisations the publication authors are affiliated to, identified using GeoNames codes (note: this field supports count: count)..',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'publication_ids': {
                    'type':
                    'string',
                    'description':
                    'The Dimensions IDs of the publications the dataset cites.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'figshare_url': {
                    'type': 'string',
                    'description': 'Figshare URL for the dataset.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'associated_publication_id': {
                    'type':
                    'string',
                    'description':
                    'The Dimensions ID of the publication linked to the dataset (single value).',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'language_title': {
                    'type':
                    'string',
                    'description':
                    'Dataset title language, as ISO 639-1 language codes.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'journal': {
                    'type': 'journals',
                    'description': 'The journal a data set belongs to.',
                    'long_description': None,
                    'is_entity': True,
                    'is_filter': True,
                    'is_facet': True
                },
                'authors': {
                    'type':
                    'json',
                    'description':
                    'Ordered list of the dataset authors. ORCIDs are included if available.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    False
                },
                'research_orgs': {
                    'type':
                    'organizations',
                    'description':
                    'GRID organisations linked to the publication associated to the dataset (note: this field supports count: count).',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'description': {
                    'type': 'string',
                    'description': 'Description of the dataset.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'researchers': {
                    'type':
                    'researchers',
                    'description':
                    "Dimensions researchers IDs associated to the dataset's associated publication. Note: in most cases, these would be the same as the dataset authors.",
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'category_icrp_cso': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Common Scientific Outline <https://app.dimensions.ai/browse/publication/cso>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'id': {
                    'type': 'string',
                    'description': 'Dimensions dataset ID.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'date_modified': {
                    'type': 'date',
                    'description':
                    'The last modification date of the dataset.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'title': {
                    'type': 'string',
                    'description': 'Title of the dataset.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': False,
                    'is_facet': False
                },
                'category_icrp_ct': {
                    'type':
                    'categories',
                    'description':
                    '`ICRP Cancer Types <https://app.dimensions.ai/browse/publication/cancer_types>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'year': {
                    'type': 'integer',
                    'description': 'Year of publication of the dataset.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                },
                'category_bra': {
                    'type':
                    'categories',
                    'description':
                    '`Broad Research Areas <https://app.dimensions.ai/browse/publication/broad_research_areas>`_',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'language_desc': {
                    'type':
                    'string',
                    'description':
                    'Dataset title language, as ISO 639-1 language codes.',
                    'long_description':
                    None,
                    'is_entity':
                    False,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'research_org_states': {
                    'type':
                    'states',
                    'description':
                    'State of the organisations the publication authors are affiliated to, expressed as GeoNames codes (ISO\u200c-3166-2).',
                    'long_description':
                    None,
                    'is_entity':
                    True,
                    'is_filter':
                    True,
                    'is_facet':
                    True
                },
                'date_created': {
                    'type': 'date',
                    'description': 'The creation date of the dataset.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': False
                },
                'repository_id': {
                    'type': 'string',
                    'description': 'The ID of the repository of the dataset.',
                    'long_description': None,
                    'is_entity': False,
                    'is_filter': True,
                    'is_facet': True
                }
            },
            'fieldsets': ['all', 'basics', 'categories'],
            'metrics': {
                'count': {
                    'name': 'count',
                    'description': 'Total count'
                }
            },
            'search_fields':
            ['full_data', 'title_only', 'title_abstract_only']
        }
    },
    'entities': {
        'categories': {
            'fields': {
                'id': {
                    'name': 'string',
                    'type': 'string',
                    'description': 'Dimensions ID of the category.',
                    'long_description': None,
                    'is_filter': True
                },
                'name': {
                    'name':
                    'string',
                    'type':
                    'string',
                    'description':
                    "Name of the category. Note: this may include an identifier from the original source. E.g., '2.1 Biological and endogenous factors' (HRCS_RAC codes) or '1103 Clinical Sciences' (FOR codes).",
                    'long_description':
                    None,
                    'is_filter':
                    True
                }
            },
            'fieldsets': ['all', 'basics']
        },
        'cities': {
            'fields': {
                'id': {
                    'name':
                    'string',
                    'type':
                    'string',
                    'description':
                    "GeoNames city ID (eg '5391811' for `geonames:5391811 <http://www.geonames.org/5391811>`_ )",
                    'long_description':
                    None,
                    'is_filter':
                    True
                },
                'name': {
                    'name': 'string',
                    'type': 'string',
                    'description': 'GeoNames city name.',
                    'long_description': None,
                    'is_filter': True
                }
            },
            'fieldsets': ['all', 'basics']
        },
        'countries': {
            'fields': {
                'id': {
                    'name':
                    'string',
                    'type':
                    'string',
                    'description':
                    "GeoNames country code (eg 'US' for `geonames:6252001 <http://www.geonames.org/6252001>`_ )",
                    'long_description':
                    None,
                    'is_filter':
                    True
                },
                'name': {
                    'name': 'string',
                    'type': 'string',
                    'description': 'GeoNames country name.',
                    'long_description': None,
                    'is_filter': True
                }
            },
            'fieldsets': ['all', 'basics']
        },
        'journals': {
            'fields': {
                'id': {
                    'name':
                    'string',
                    'type':
                    'string',
                    'description':
                    'Dimensions ID of a journal. E.g., `jour.1016355 <https://app.dimensions.ai/discover/publication?and_facet_source_title=jour.1016355>`_ or `jour.1077219 <https://app.dimensions.ai/discover/publication?and_facet_source_title=jour.1077219>`_ .',
                    'long_description':
                    None,
                    'is_filter':
                    True
                },
                'title': {
                    'name':
                    'string',
                    'type':
                    'string',
                    'description':
                    "Title of a journal publication. E.g. 'Nature' or 'The Lancet'",
                    'long_description':
                    None,
                    'is_filter':
                    True
                }
            },
            'fieldsets': ['all', 'basics']
        },
        'org_groups': {
            'fields': {
                'id': {
                    'name': 'string',
                    'type': 'string',
                    'description': 'Dimensions ID of the organization group.',
                    'long_description': None,
                    'is_filter': True
                },
                'name': {
                    'name':
                    'string',
                    'type':
                    'string',
                    'description':
                    "Name of the organization group. E.g., 'NIH' or 'ICRP'.",
                    'long_description':
                    None,
                    'is_filter':
                    True
                }
            },
            'fieldsets': ['all', 'basics']
        },
        'states': {
            'fields': {
                'id': {
                    'name':
                    'string',
                    'type':
                    'string',
                    'description':
                    "GeoNames state code (ISO\u200c-3166-2). E.g., 'US.CA' for `geonames:5332921 <http://www.geonames.org/5332921>`_ .",
                    'long_description':
                    None,
                    'is_filter':
                    True
                },
                'name': {
                    'name': 'string',
                    'type': 'string',
                    'description': 'GeoNames state name (ISO\u200c-3166-2).',
                    'long_description': None,
                    'is_filter': True
                }
            },
            'fieldsets': ['all', 'basics']
        },
        'open_access': {
            'fields': {
                'id': {
                    'name':
                    'string',
                    'type':
                    'string',
                    'description':
                    "Dimensions ID of the open access category. E.g., one of 'closed', 'oa_all', 'gold_bronze', 'gold_pure', 'green_sub', 'gold_hybrid', 'green_pub', 'green_acc'. (see also the :ref:`publications` field ``open_access``).",
                    'long_description':
                    None,
                    'is_filter':
                    True
                },
                'name': {
                    'name':
                    'string',
                    'type':
                    'string',
                    'description':
                    "Name of the open access category. E.g., 'Closed' or 'Pure Gold'.",
                    'long_description':
                    None,
                    'is_filter':
                    True
                },
                'description': {
                    'name': 'string',
                    'type': 'string',
                    'description': 'Description of the open access category.',
                    'long_description': None,
                    'is_filter': False
                }
            },
            'fieldsets': ['all', 'basics']
        }
    }
}



# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Unit tests for Dimcli 
"""

from __future__ import print_function

import unittest, os, sys, click
import configparser

from .. import *
from ..core.auth import USER_CONFIG_FILE_PATH
from ..core.utils import dimensions_url
from ..core.api import DslDataset
from ..shortcuts import *


class TestOne(unittest.TestCase):

    """
    Tests - iterative queries and other related features 
    """

    click.secho("**TESTS**", fg="red")
    login(instance="live")
    d = Dsl()

    def test_001(self):
        click.secho("\nTEST 001: Iterative querying.", fg="green")
        # ----
        d = Dsl()
        res = d.query_iterative("""search publications where journal.title="nature medicine" and year>2000 return publications""")
        click.secho("Query results: ", fg="magenta")
        print(" ==> len(res): ", len(res))
        print(" ==> res['stats']: ", res['stats'])
        print(" ==> len(res['publications']): ", len(res['publications']))
        # ----
        click.secho("\nTEST 001: Iterative querying with custom pause: 5 seconds.", fg="green")
        # ----
        d = Dsl()
        res = d.query_iterative("""search publications where journal.title="nature medicine" and year>2015 return publications""", pause=5)
        click.secho("Query results: ", fg="magenta")
        print(" ==> len(res): ", len(res))
        print(" ==> res['stats']: ", res['stats'])
        print(" ==> len(res['publications']): ", len(res['publications']))
        # ----
        click.secho("\nTEST 001: Iterative querying with force=True", fg="green")
        # ----
        d = Dsl()
        q = """search publications 
        where research_orgs.id="grid.5522.0" 
        and year in [2010:2012]
        return publications[basics+category_for+times_cited]
        """
        res = d.query_iterative(q, limit=1000, force=True)
        click.secho("Query results: ", fg="magenta")
        print(" ==> len(res): ", len(res))
        print(" ==> res['stats']: ", res['stats'])
        print(" ==> len(res['publications']): ", len(res['publications']))
        # ----
        click.secho("Completed test succesfully", fg="green")


    def test_002(self):
        click.secho("\nTEST 002: Query Shortcuts", fg="green")
        
        # ----
        res = dslquery("""search publications where journal.title="nature medicine" return publications limit 10""")
        click.secho("Query results for `dslquery`: ", fg="magenta")
        print(" ==> res['stats']: ", res['stats'])
        print(" ==> len(res['publications']): ", len(res['publications']))
        # ----
        res = dslquery_json("""search publications where journal.title="nature medicine" return publications limit 10""")
        click.secho("Query results for `dslquery_json`: ", fg="magenta")
        print(" ==> type(res): ", type(res))
        # ----
        res = dslqueryall("""search publications where year="1815" return publications""")
        click.secho("Query results for `dslqueryall`: ", fg="magenta")
        print(" ==> type(res): ", type(res))
        # ----
        click.secho("Completed test succesfully", fg="green")

    def test_003(self):
        click.secho("\nTEST 003: Non-search queries.", fg="green")
        # ----
        res= dslquery("""extract_grants(grant_number="185247", funder_name="Swiss National Science Foundation")""")
        click.secho("Query results: ", fg="magenta")
        print(" ==> res: ", res)
        print(" ==> res.json: ", res.json)
        # ----
        click.secho("Completed test succesfully", fg="green")


    def test_004(self):
        click.secho("\nTEST 004: Try magic methods on DslDataset object.", fg="green")
        # ----
        login(instance="live")
        d = Dsl()
        click.secho("Query #1... returning publications", fg="green")
        res = d.query("search publications where year=2018 return publications")
        print("Query results: ")
        print(" ==> res.json.keys(): ", res.json.keys())
        print(" ==> res['publications'][0]: ", res['publications'][0])
        print(" ==> len(res): ", len(res))
        print(" ==> res['stats']: ", res['stats'])
        print(" ==> res['not_there']: ", res['not_there'])
        print(" ==> res.publications[0]: ", res.publications[0])
        print(" ==> res.stats: ", res.stats)
        # ----
        click.secho("Query #2... returning facet", fg="green")
        res = d.query("""search publications for \"bmw\" return year""")
        print("Query results: ")
        print(" ==> res.json.keys(): ", res.json.keys())
        print(" ==> res['year'][0]: ", res['year'][0])
        print(" ==> len(res): ", len(res))
        print(" ==> res['stats']: ", res['stats'])
        print(" ==> res['not_there']: ", res['not_there'])
        print(" ==> res.year[0]: ", res.year[0])
        print(" ==> res.stats: ", res.stats)
        click.secho("Completed test succesfully", fg="green")


    def test_005(self):
        click.secho("\nTEST 005: Chunking Results", fg="green")
        
        # ----
        res = dslquery("""search publications where journal.title="nature medicine" return publications limit 1000""")
        click.secho("Query results for standard query: ", fg="magenta")
        print(" ==> res['stats']: ", res['stats'])
        print(" ==> len(res['publications']): ", len(res['publications']))
        click.secho("Now chunking with default size: ", fg="magenta")
        test = [len(x) for x in res.chunks()]
        print(" ==> [len(x) for x in res.chunks()]: ", str(test)) 
        click.secho("Now chunking with size 50: ", fg="magenta")
        test = [len(x) for x in res.chunks(50)]
        print(" ==> [len(x) for x in res.chunks()]: ", str(test)) 
        click.secho("Now chunking with wrong key (should fail): ", fg="magenta")
        test = [len(x) for x in res.chunks(key="badkey")]
        # ----

    def test_006(self):
        click.secho("\nTEST 006: Building dimcli.DslDataset objects from data.", fg="green")
        # ----
        data = dslquery("""search publications for "malaria" return publications""")
        res = DslDataset.from_publications_list(data.publications)
        res.as_dataframe_authors()
        res.count_total
        data = dslquery("""search patents for "graphene" return patents""")
        res = DslDataset.from_patents_list(data.patents)
        res.as_dataframe()
        res.count_total
        # ----
        click.secho("\nBuilding dimcli.DslDataset objects from DATAFRAME data.", fg="green")
        # ----
        data = dslquery("""search publications for "malaria" return publications""").as_dataframe()
        res = DslDataset.from_publications_list(data)
        res.as_dataframe_authors()
        res.count_total
        data = dslquery("""search patents for "graphene" return patents""").as_dataframe()
        res = DslDataset.from_patents_list(data)
        res.as_dataframe()
        res.count_total
        click.secho("Completed test succesfully", fg="green")


if __name__ == "__main__":
    unittest.main()




from IPython.core import magic_arguments
from IPython.core.magic import line_magic, cell_magic, line_cell_magic, Magics, magics_class

from ..VERSION import VERSION

from ..core.api import Dsl
from ..core.auth import get_connection, is_logged_in
from ..core.utils import *


@magics_class
class DslMagics(Magics):

    results_var = "dsl_last_results" # VAR AUTOMATICALLY SET TO LATEST QUERY RESULTS
    dslobject = None


    def _handle_login(self):
        if is_logged_in():
            self.dslobject = Dsl(show_results=False, verbose=True)
            return True
        else:
            return False


    def _handle_query(self, text, loop=False):
        """main procedure after user input"""
       
        # lazy complete
        text = line_add_lazy_return(text)
        text = line_add_lazy_describe(text)

        # RUN QUERY
        if not loop:
            res = self.dslobject.query(text)
            return res
        else:
            res = self.dslobject.query_iterative(text)
            return res
        

    #
    # MAGIC METHODS
    # 


    @line_cell_magic
    def dsl(self, line, cell=None):
        """
        Dimcli Magic Method: enter a DSL query string and send it directly to Dimensions API. 
        """
        if self._handle_login():
            if cell:
                line = cell
            data = self._handle_query(line)
            self.shell.user_ns[self.results_var] = data
            return data


    @line_cell_magic
    def dsldf(self, line, cell=None):
        """Dimcli Magic method: query the Dimensions DSL API with the text passed - return a pandas dataframe.
        """
        if self._handle_login():
            if cell:
                line = cell
            if not line_is_search_query(line):
                print("Sorry - DSL to dataframe magic methods work only with `search` queries.")
                return None
            data = self._handle_query(line).as_dataframe()
            self.shell.user_ns[self.results_var] = data
            return data


    @line_cell_magic
    def dslloop(self, line, cell=None):
        """Dimcli Magic Method: transforms a simple query into a loop, by adding limit/skip parameters automatically. The final object returns contains all results into a single JSON object.  
        """
        if self._handle_login():
            if cell:
                line = cell
            data = self._handle_query(line, loop=True)
            self.shell.user_ns[self.results_var] = data
            return data


    @line_cell_magic
    def dslloopdf(self, line, cell=None):
        """Dimcli Magic method: query the Dimensions DSL API with the text passed, looping over all results pages - return a pandas dataframe.
        """
        if self._handle_login():
            if cell:
                line = cell
            if not line_is_search_query(line):
                print("Sorry - DSL to dataframe magic methods work only with `search` queries.")
                return None
            data = self._handle_query(line, loop=True).as_dataframe()
            self.shell.user_ns[self.results_var] = data
            return data


    @line_magic
    def dsldocs(self, line):
        """
        Dimcli Magic method: wrapper around the dsl `describe` function - outputs documentation for a DSL Source or Entity as a pandas dataframe.
        """
        if not self._handle_login():
            return

        try:
            import pandas as pd
            df = pd.DataFrame()
        except:
            print("Sorry this functionality requires the Pandas python library. Please install it first")
            return

        obj = line.strip()
        if obj and obj not in G.entities() and obj not in G.sources():
            sou = " - ".join([x for x in G.sources()])
            ent = " - ".join([x for x in G.entities()])
            print(f"Can't recognize this object. Dimcli knows about:\n Sources=[{sou}] Entities=[{ent}] ")
            # continue anyways
        
        res = self._handle_query(f"describe schema") # same query for all requests (filtering done here)

        if not obj:
            # show data for all sources
            docs_for = G.sources()
            header = "sources"
        elif obj in G.entities():
            # match single entity
            docs_for = [obj]
            header = "entities"
        elif obj in G.sources():
            # match single source
            docs_for = [obj]
            header = "sources"
        else:
            # unknown key: try to match a source (eg for newly published sources)
            docs_for = [obj]
            header = "sources"            

        d = {header: [], 'field': [], 'type': [], 'description':[], 'is_filter':[], 'is_entity': [],  'is_facet':[],}
        for S in docs_for:
            for x in sorted(res.json[header][S]['fields']):
                d[header] += [S]
                d['field'] += [x]
                d['type'] += [res.json[header][S]['fields'][x]['type']]
                d['description'] += [res.json[header][S]['fields'][x]['description']]
                d['is_filter'] += [res.json[header][S]['fields'][x]['is_filter']]
                d['is_facet'] += [res.json[header][S]['fields'][x].get('is_facet', False)]
                d['is_entity'] += [res.json[header][S]['fields'][x].get('is_entity', False)]

        data = df.from_dict(d)
        self.shell.user_ns[self.results_var] = data
        return data








ip = get_ipython()
ip.register_magics(DslMagics)

# https://github.com/ipython/ipython/issues/11878#issuecomment-554790961
ip.Completer.use_jedi = False

from ..repl.autocompletion import CleverCompleter
from prompt_toolkit.document import Document

def load_ipython_custom_completers(ipython):
    """

    # TODO 
    # check more Completer behaviour in order to remove extra suggestions eg
    # https://www.programcreek.com/python/example/50972/IPython.get_ipython


    """

    def dslq_completers(self, event):
        """ This should return a list of strings with possible completions.

        Note that all the included strings that don't start with event.symbol
        are removed, in order to not confuse readline.

        eg Typing %%apt foo then hitting tab would yield an event like so: namespace(command='%%apt', line='%%apt foo', symbol='foo', text_until_cursor='%%apt foo')

        https://stackoverflow.com/questions/36479197/ipython-custom-tab-completion-for-user-magic-function

        > https://github.com/ipython/ipython/issues/11878
        """
        # print(dir(event), event)

        # FIXME only first line gets the autocomplete!
        if event.line.startswith("%%"):
            event.line = event.line[1:] #  reduce cell symbol (double %) to line symbol
        for command in ["%dslloopdf", "%dsldf", "%dslloop", "%dsl"]:
            if command in event.line: # first match will return results
                doc = Document(event.line.replace(command, ""))
                c = CleverCompleter()
                res = c.get_completions(doc, None)
                # print(res)
                return [x.text for x in res]
            
    def dsldocs_completers(self, event):
        """ 
        Completer for dsldocs command (= describe)
        """
        # print(dir(event), event)
        command = "%dsldocs"
        if event.line.startswith(command):
            doc = Document(event.line.replace(command, ".docs"))
            c = CleverCompleter()
            res = c.get_completions(doc, None)
            # print(res)
            return [x.text for x in res]           

    
    # loader

    for command in ["%dslloop_to_dataframe", "%dsl_to_dataframe", "%dslloop", "%dsl"]:
        ipython.set_hook('complete_command', dslq_completers, re_key = command)
        ipython.set_hook('complete_command', dslq_completers, re_key = "%" + command)

    ipython.set_hook('complete_command', dsldocs_completers, re_key = "%dsldocs")


load_ipython_custom_completers(ip)




# ****************************************************************************

# 
# MEMOS 
# 

# ARGUMENTS AND LINE MAGICS 
# 
# line magics and arguments don't play well together when you need to parse a line content
# in that case the args parsing throws an error
# hence better to have a cell magic, and pass only the args in the line

# @magic_arguments.magic_arguments()
# @magic_arguments.argument('--endpoint',
#     help='The query endpoint: default is https://app.dimensions.ai'
# )
# @magic_arguments.argument('--user',
#     help='The account username'
# )
# @magic_arguments.argument('--password',
#     help='The account password'
# )
# @magic_arguments.argument('--env',
#     help='The instance name as defined in init.py. Default: live.'
# )


# ****************************************************************************


"""
Module for quick tests and experiemnts. 

None of which is final in any way. But could become some day..
"""

try:
    from pyvis.network import Network as PyvisNetwork
except ImportError:
    print("ERROR: this module requires pyvis `pip install pyvis`")
    raise

import IPython
import os
from .. import COLAB_ENV


class NetworkViz(PyvisNetwork):
    """Extend PyVis class so that we can use a modified template. 

    Note: this is an experimental feature. 
    The modified template allows to display JS inline hence it works in Colab. 
    """
 
    def __init__(self,
                 height="500px",
                 width="500px",
                 directed=False,
                 notebook=False,
                 bgcolor="#ffffff",
                 font_color=False,
                 layout=None):
        # call super class init
        PyvisNetwork.__init__(self, 
                              height, 
                              width,
                              directed,
                              notebook,
                              bgcolor,
                              font_color,
                              layout)
        # override template location
        self.path =  os.path.abspath(os.path.join(os.path.dirname( __file__ ), '..', 'templates/pyvis_inline.html'))
        # self.path = os.path.dirname(__file__) + "/templates/pyvis_inline.html"

    def check_html(self, name):
        """
        Given a name of graph to save or write, check if it is of valid syntax

        :param: name: the name to check
        :type name: str
        """
        assert len(name.split(".")) >= 2, "invalid file type for %s" % name
        assert name.split(
            ".")[-1] == "html", "%s is not a valid html file" % name


    def show(self, name):
        """
        Writes a static HTML file and saves it locally before opening.

        :param: name: the name of the html file to save as
        :type name: str
        """
        self.check_html(name)
        if self.template is not None:
            if not COLAB_ENV: 
                # return IFrame(name, width=self.width, height=self.height)
                return self.write_html(name, notebook=True)
            else:
                self.write_html(name, notebook=True)
                return IPython.display.HTML(filename=name)
        else:
            self.write_html(name)
            webbrowser.open(name)


# https://gist.github.com/zdavkeos/1098474

import os
from os import path

def walk_up(bottom):
    """ 
    mimic os.walk, but walk 'up'
    instead of down the directory tree
    """

    bottom = path.realpath(bottom)

    #get files in current dir
    try:
        names = os.listdir(bottom)
    except Exception as e:
        print(e)
        return


    dirs, nondirs = [], []
    for name in names:
        if path.isdir(path.join(bottom, name)):
            dirs.append(name)
        else:
            nondirs.append(name)

    yield bottom, dirs, nondirs

    new_path = path.realpath(path.join(bottom, '..'))
    
    # see if we are at the top
    if new_path == bottom:
        return

    for x in walk_up(new_path):
        yield x


if __name__ == '__main__':
    # tests/demos

    #print all files and directories
    # directly above the current one
    for i in walk_up(os.curdir):
        print(i)

    # look for a TAGS file above the
    # current directory
    for c,d,f in walk_up(os.curdir):
        if 'TAGS' in f:
            print(c)
            break
